/*
    $Id: covariance_decomposition.h 246 2011-09-01 07:18:57Z matwey $
    Copyright (C) 2011  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _COVARIANCE_DECOMPOSITION_H
#define _COVARIANCE_DECOMPOSITION_H

#include <cmath>

#include <boost/format.hpp>
#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/tuple/tuple.hpp>

#include <lsp/cholesky_decomposition.h>
#include <lsp/schur_decomposition.h>

#include "ublas_math.h"

namespace utils {
namespace ublas {

using namespace boost::numeric::ublas;

namespace detail {
	template<class M> boost::tuple< banded_matrix< typename M::value_type >, symmetric_matrix< typename M::value_type > > correlation_matrix( const M& cov ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		symmetric_matrix< value_type > cor(cov);
		banded_matrix< value_type > sigma = utils::ublas::sqrt( banded_adaptor< symmetric_matrix< value_type > >(cor) );
		for( typename banded_matrix< value_type >::size_type i = 0; i < sigma.size1(); ++i ) {
			row(cor,i)    /= sigma(i,i);
			column(cor,i) /= sigma(i,i);
		}

		return boost::make_tuple( sigma, cor );
	}
	template<class M> triangular_matrix< typename M::value_type > sqrt_matrix_decomposition( const M& cov ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		symmetric_matrix< value_type > decomp(cov);
		lsp::cholesky_decomposition( decomp );
		triangular_adaptor< symmetric_matrix< value_type > > result(decomp);
		
		return result;
	}
	template<class M> triangular_matrix< typename M::value_type > inv_sqrt_matrix_decomposition( const M& cov ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		triangular_matrix< value_type > triang( sqrt_matrix_decomposition(cov) );
		matrix< value_type > inv = solve( triang, identity_matrix< value_type >( triang.size1() ), lower_tag() );

		return triangular_adaptor< matrix< value_type > >(inv);
	}
	template<class M> triangular_matrix< typename M::value_type > aetkin_weight_matrix( const M& cov, const typename M::value_type& cond_threshold = 10000 ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		if( BOOST_UBLAS_SAME( cov.size1(), cov.size2() ) < 1 )
			return triangular_matrix< value_type >(0,0);

		for( size_type i = 0; i < BOOST_UBLAS_SAME( cov.size1(), cov.size2() ); ++i ) {
			if( cov(i,i) <= value_type(0) )
				throw std::logic_error("Non-positive diagonal value");
		}

		boost::tuple< banded_matrix< value_type >, symmetric_matrix< value_type > > sigma_cor = correlation_matrix(cov);
		triangular_matrix< value_type > weight = sqrt_matrix_decomposition( sigma_cor.template get<1>() );

		value_type min = std::abs( weight(0,0) );
		value_type max = std::abs( weight(0,0) );
		for( typename triangular_matrix< value_type >::size_type i = 1; i < weight.size1(); ++i ) {
			if( !std::isfinite(weight(i,i)) ) throw std::logic_error( "Infinite diagonal value" );
			if( std::abs( weight(i,i) ) > max ) max = std::abs( weight(i,i) );
			if( std::abs( weight(i,i) ) < min ) min = std::abs( weight(i,i) );
		}

		value_type cond = max / min;
		if( !std::isfinite( cond ) || cond > cond_threshold ) {
			throw std::logic_error( (boost::format("The condition number of covariance matrix is %1%") % cond).str() );
		}

		matrix< value_type > inv = solve( weight, identity_matrix< value_type >( weight.size1() ), lower_tag() );
		inv = prod( inv, utils::ublas::inv( sigma_cor.template get<0>() ) );
		return triangular_adaptor< matrix< value_type > >(inv);
	}
	// FIXME: result is a symmetric matrix. U L^1/2 U^T
	template<class M> std::pair< matrix< typename M::value_type >, matrix< typename M::value_type > > eigen_weight_matrix( const M& cov, const typename M::value_type& cond_threshold = 100000 ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;

		if( BOOST_UBLAS_SAME(cov.size1(),cov.size2()) < 1 )
			return std::make_pair( matrix< value_type >(0,0), matrix< value_type >(0,0) );
		
		symmetric_matrix< value_type > lambda = cov;
		lsp::schur_decomposition< M > sd(lambda);
		matrix< value_type > left = identity_matrix< value_type >(BOOST_UBLAS_SAME(cov.size1(),cov.size2()));
		sd.apply(left);
	
		banded_adaptor< symmetric_matrix< value_type > > diag_lambda(lambda);
		banded_matrix< value_type > diag_invlambda(lambda.size1(),lambda.size2());
		
		value_type max_lambda = std::abs(diag_lambda(0,0));
		for( size_type i = 0; i < BOOST_UBLAS_SAME(diag_lambda.size1(),diag_lambda.size2()); ++i ) {
			if( cond_threshold*diag_lambda(i,i) < max_lambda || diag_lambda(i,i) < 0 ) {
				diag_invlambda(i,i) = 0;
				diag_lambda(i,i) = 0;
			} else {
				diag_invlambda(i,i) = value_type(1.0)/std::sqrt(diag_lambda(i,i));
				diag_lambda(i,i) = std::sqrt(diag_lambda(i,i));
			}
		}
		
		return std::make_pair(
			prod(trans(left),matrix<value_type>(prod(diag_invlambda,left))),
			prod(trans(left),matrix<value_type>(prod(diag_lambda,left))) );
	}
	template<class M> std::pair< banded_matrix< typename M::value_type >, banded_matrix< typename M::value_type > > gauss_weight_matrix( const M& cov, const typename M::value_type& cond_threshold = 10000  ) {
		typedef M matrix_type;
		typedef typename M::value_type value_type;
		typedef typename M::size_type  size_type;
		
		if( BOOST_UBLAS_SAME( cov.size1(), cov.size2() ) < 1 )
			return std::make_pair( banded_matrix< value_type >(0,0), banded_matrix< value_type >(0,0) );

		for( size_type i = 0; i < BOOST_UBLAS_SAME(cov.size1(),cov.size2()); ++i ) {
			if( cov(i,i) <= value_type(0) )
				throw std::logic_error("Non-positive diagonal value");
		}
		
		value_type min = std::abs( cov(0,0) );
		value_type max = std::abs( cov(0,0) );
		for( size_type i = 1; i < BOOST_UBLAS_SAME(cov.size1(),cov.size2()); ++i ) {
			if( !std::isfinite(cov(i,i)) ) throw std::logic_error( "Infinite diagonal value" );
			if( std::abs( cov(i,i) ) > max ) max = std::abs( cov(i,i) );
			if( std::abs( cov(i,i) ) < min ) min = std::abs( cov(i,i) );
		}

		value_type cond = max / min;
		if( !std::isfinite( cond ) || cond > cond_threshold ) {
			throw std::logic_error( (boost::format("The condition number of covariance matrix is %1%") % cond).str() );
		}
		
		banded_matrix< value_type > ret(cov.size1(),cov.size2());
		banded_matrix< value_type > invret(cov.size1(),cov.size2());
		
		for( size_type i = 0; i < BOOST_UBLAS_SAME(cov.size1(),cov.size2()); ++i ) {
			ret(i,i) = value_type(1)/std::sqrt(cov(i,i));
			invret(i,i) = std::sqrt(cov(i,i));
		}
		
		return std::make_pair( ret, invret );
	}
}

template<class T> boost::tuple< banded_matrix<T>, symmetric_matrix<T> > correlation_matrix( const symmetric_matrix<T>& cov ) {
	return detail::correlation_matrix( cov );
}
template<class T> triangular_matrix<T> sqrt_matrix_decomposition( const symmetric_matrix<T>& cov ) {
	return detail::sqrt_matrix_decomposition( cov );
}
template<class T> triangular_matrix<T> inv_sqrt_matrix_decomposition( const symmetric_matrix<T>& cov ) {
	return detail::inv_sqrt_matrix_decomposition( cov );
}
template<class T> triangular_matrix<T> aetkin_weight_matrix( const symmetric_matrix<T>& cov ) {
	return detail::aetkin_weight_matrix( cov );
}
template<class T> std::pair< matrix<T>, matrix<T> > eigen_weight_matrix( const symmetric_matrix<T>& cov ) {
	return detail::eigen_weight_matrix( cov );
}
template<class T> std::pair< banded_matrix<T>, banded_matrix<T> > gauss_weight_matrix( const banded_matrix<T>& cov ) {
	return detail::gauss_weight_matrix( cov );
}

template<class M> boost::tuple< banded_matrix< typename M::value_type >, symmetric_matrix< typename M::value_type > > correlation_matrix( const symmetric_adaptor<M>& cov ) {
	return detail::correlation_matrix( cov );
}
template<class M> triangular_matrix< typename M::value_type > sqrt_matrix_decomposition( const symmetric_adaptor<M>& cov ) {
	return detail::sqrt_matrix_decomposition( cov );
}
template<class M> triangular_matrix< typename M::value_type > inv_sqrt_matrix_decomposition( const symmetric_adaptor<M>& cov ) {
	return detail::inv_sqrt_matrix_decomposition( cov );
}
template<class M> triangular_matrix< typename M::value_type > aetkin_weight_matrix( const symmetric_adaptor<M>& cov ) {
	return detail::aetkin_weight_matrix( cov );
}
template<class M> std::pair< matrix< typename M::value_type >, matrix< typename M::value_type > > eigen_weight_matrix( const symmetric_adaptor<M>& cov ) {
	return detail::eigen_weight_matrix( cov );
}
template<class M> std::pair< banded_matrix< typename M::value_type >, banded_matrix< typename M::value_type > > gauss_weight_matrix( const banded_adaptor<M>& cov ) {
	return detail::gauss_weight_matrix( cov );
}

}}

#endif // _COVARIANCE_DECOMPOSITION_H
