/*
    $Id: profile.cpp 246 2011-09-01 07:18:57Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <lsp/nnls.h>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <numeric>

#include "atmos.h"
#include "io.h"
#include "profile.h"
#include "ublas_math.h"
#include "utils.h"

namespace atmos {
namespace mixture {

storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_mixture_column( *this, boost::bind(&detail::sample::mixture, _1) ) {
}
storage::storage( const atmos::indices::storage& indices, const atmos::dimm::storage& dimm, double dimmscale ):
	table_storage< detail::sample >( indices.size() ),
	m_mixture_column( *this, boost::bind(&detail::sample::mixture, _1) ) {
	using namespace boost::numeric::ublas;
	
	if( dimm.size() > 1 ) {
		double dimmscale2 = dimmscale*dimmscale;
		dimm::storage::const_iterator dimm_current = dimm.begin();
		dimm::storage::const_iterator dimm_next = dimm_current + 1;
		indices::storage::const_iterator it = indices.begin();

		while( dimm_current != dimm.end() ) {
			while( it != indices.end() ) {
				if( dimm_next != dimm.end() && dimm_next->timestamp <= it->timestamp )
					break;

				mixture_type mixture( atmos::indices_size + atmos::dimm_indices_size );
				project( mixture, range(0,atmos::indices_size) ) = atmos::indices::make_atmos_order(it->indices);
				mixture(10) = dimmscale2 * dimm_current->xvars(1) * 1e+11;
				mixture(11) = dimmscale2 * dimm_current->yvars(1) * 1e+11;
				++it;
				push_back( it->timestamp, mixture );
			}
			dimm_current = dimm_next;
			++dimm_next;
		}
	} else {
		for( indices::storage::const_iterator it = indices.begin(); it != indices.end(); ++it ) {
			push_back( it->timestamp, atmos::indices::make_atmos_order(it->indices) );
		}
	}
}

}
namespace profile {

const double MAXCHI2 = 999.0;
const double JPOW_SEE = 0.6 ;
const double LPOW_SEE = -0.2;
const double KL = 1.1486983549970351; // 0.5 ** LPOW_SEE
const double KS = 1.73588e+07;

using namespace boost::numeric::ublas;

namespace detail {

//FIXME:
template<class VE> typename VE::size_type nan_elements( const vector_expression<VE>&x ) {
	typename VE::size_type n = 0;
	for( typename VE::size_type j = 0; j < x().size(); j++ )
		n += x()(j) != x()(j) ? 1 : 0;
	return n;
}

}

functor::functor( const grid_type& grid, const mixture_covariance_type& cov_matrix, double secz, const weight_type& basewf ):
	m_grid( grid ), m_cov_matrix(cov_matrix), m_secz(secz) {
	using namespace utils::ublas;
	using namespace std;

	const value_type cond_numer_threshold = 10000;

	banded_matrix< value_type > sigma = sqrt( banded_adaptor< mixture_covariance_type >( m_cov_matrix ) );
	for( banded_matrix< value_type >::size_type i = 0; i < sigma.size1(); ++i ) {
		row(m_cov_matrix,i)    /= sigma(i,i);
		column(m_cov_matrix,i) /= sigma(i,i);
	}

	lsp::cholesky_decomposition( m_cov_matrix, upper_tag() );
	triangular_adaptor< mixture_covariance_type, lower > R( m_cov_matrix );

	vector_type abs_diag = abs( diag( R ) );
	value_type cond_number = ( detail::nan_elements(abs_diag) ? std::numeric_limits< value_type >::quiet_NaN() : *max_element(abs_diag.begin(),abs_diag.end()) / *min_element(abs_diag.begin(),abs_diag.end()) );

	if( !std::isfinite( cond_number ) || cond_number > cond_numer_threshold ) {
		throw std::logic_error( (boost::format("The condition number of covariance matrix is %1%") % cond_number).str() );
	}

	m_weight_matrix = identity_matrix< value_type >( m_cov_matrix.size1() );
	inplace_solve( R, m_weight_matrix, lower_tag() );
	m_weight_matrix = prod( m_weight_matrix, inv(sigma) );
	R = prod( sigma, R );

	m_problem_matrix = prod( m_weight_matrix, subrange( vector_of_vector_to_matrix( project(grid*secz, basewf) ), 0, m_weight_matrix.size2(),0,grid.size()) );
}

functor::result_type functor::operator() ( const mixture_type& indices ) const {
	vector_type weighted_indices = prod( m_weight_matrix, indices );

	result_type ret;

	matrix_type cov = zero_matrix<double>( m_problem_matrix.size2(), m_problem_matrix.size2() );

	lsp::nnls< matrix_type, vector_type > nnls( m_problem_matrix, weighted_indices );
	nnls.solve( ret.get<0>(), cov );

	ret.get<1>() = prod( m_problem_matrix, ret.get<0>() ) - weighted_indices;

	ret.get<2>() = std::min( inner_prod( ret.get<1>(), ret.get<1>() ), MAXCHI2 );

	ret.get<0>() /= m_secz;
	return ret;
}
double functor::seeing( const cn2_type& cn2 ) const {
	return KS*KL*std::pow( sum( cn2 ), JPOW_SEE );
}
vector_type functor::indices_residual( const residual_type& residual ) const {
	return prod( triangular_adaptor< const matrix< double >, lower >(m_cov_matrix), residual );
}

storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_cn2_column( *this, boost::bind(&detail::sample::cn2, _1) ),
	m_residual_column( *this, boost::bind(&detail::sample::residual, _1)  ),
	m_chi2_column( *this, boost::bind(&detail::sample::chi2, _1) ) {
}

namespace filter {

chi2_gt_than::chi2_gt_than( const value_type& chi2 ): chi2_(chi2) {
}
bool chi2_gt_than::operator()( const sample_type& x ) const {
	return x.chi2 > chi2_;
}

}
}}