/*
    $Id: variableexposure_profile.cpp 247 2011-09-02 06:14:33Z matwey $
    Copyright (C) 2011  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/function_output_iterator.hpp>
#include <boost/mem_fn.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>


#include <steihaug_toint.h>

#include "atmos.h"
#include "covariance_decomposition.h"
#include "merge_join.h"
#include "variableexposure_profile.h"
#include "ublas_math.h"
#include "utils.h"

namespace atmos {
namespace variableexposure {
namespace initial {

namespace detail {
struct pack_input {
	typedef void result_type;
	typedef std::pair< atmos::profile::storage::value_type, atmos::shortexposure::profile::storage::value_type > arg_type;

	pack_input( storage& storage ): storage_(storage) { }
	void operator() ( const arg_type& x ) const {
		using namespace utils::ublas;

		storage_.push_back( x.first.timestamp, x.first.cn2, sqrt( element_div( x.second.windspeed, x.first.cn2 ) ) );
	}
private:
	storage& storage_;
};
}
	
storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_cn2_column(*this, boost::mem_fn(&detail::sample::cn2) ),
	m_windspeed_column(*this, boost::mem_fn(&detail::sample::windspeed) ) {

}
storage::storage( const atmos::profile::storage& profile, const atmos::shortexposure::profile::storage& windspeed ):
	table_storage< detail::sample >( std::min( profile.size(), windspeed.size() ) ),
	m_cn2_column(*this, boost::mem_fn(&detail::sample::cn2) ),
	m_windspeed_column(*this, boost::mem_fn(&detail::sample::windspeed) ) {
	using utils::algo::merge_join;

	merge_join( profile.begin(), profile.end(), windspeed.begin(), windspeed.end(),
		    boost::make_function_output_iterator( detail::pack_input(*this) ),
		    boost::mem_fn( &atmos::profile::storage::value_type::timestamp),
		    boost::mem_fn( &atmos::shortexposure::profile::storage::value_type::timestamp) );
}

void storage::repair_initial_cn2( const cn2_type& avg_cn2 ) {
//FIXME: must be a standalone function when ported
	for( iterator it = begin(); it != end(); ++it ) {
		// we dont want zero turbulence as initial value
		value_type::cn2_type::value_type oldnorm = norm_1(it->cn2);
		for( value_type::cn2_type::size_type i = 0; i < it->cn2.size(); ++i ) {
			if( it->cn2(i) == value_type::cn2_type::value_type(0) ) {
				if( avg_cn2(i) != value_type::cn2_type::value_type(0) ) {
					it->cn2(i) = avg_cn2(i);
				} else {
					it->cn2(i) = oldnorm / it->cn2.size();
				}
			}
		}
		value_type::cn2_type::value_type newnorm = norm_1(it->cn2);
		it->cn2 = it->cn2 * oldnorm / newnorm;
	}
}
void storage::repair_initial_windspeed( const windspeed_type& avg_windspeed ) {
//FIXME: must be a standalone function when ported
	for( iterator it = begin(); it != end(); ++it ) {
		// we dont want zero of undefined wind as initial value
		value_type::windspeed_type::value_type oldnorm = norm_1(it->windspeed);
		for( value_type::windspeed_type::size_type i = 0; i < it->windspeed.size(); ++i ) {
			if( !std::isfinite(it->windspeed(i)) || it->windspeed(i) == value_type::windspeed_type::value_type(0) ) {
				if( std::isfinite(avg_windspeed(i)) && avg_windspeed(i) != value_type::windspeed_type::value_type(0) ) {
					it->windspeed(i) = avg_windspeed(i);
				} else {
					it->windspeed(i) = value_type::windspeed_type::value_type(1);
				}
			}
		}
	}
}

} // initial

namespace input {
	
namespace {
struct pack_input {
	typedef void result_type;
	typedef std::pair< atmos::indices::multiexp::storage::value_type, atmos::variableexposure::initial::storage::value_type > arg_type;

	pack_input( storage& storage ): storage_(storage) { }
	void operator() ( const arg_type& x ) const {
		using namespace utils::ublas;
		
		storage_.push_back( x.first.timestamp, atmos::indices::multiexp::make_atmos_order(x.first.indices), x.second.cn2, x.second.windspeed );
	}
private:
	storage& storage_;
};
}
	
storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_indices_column(*this, boost::mem_fn(&detail::sample::indices) ),
	m_cn2_column(*this, boost::mem_fn(&detail::sample::cn2) ),
	m_windspeed_column(*this, boost::mem_fn(&detail::sample::windspeed) ) {
}
storage::storage( const atmos::indices::multiexp::storage& indices, const atmos::variableexposure::initial::storage& initial ):
	table_storage< detail::sample >( std::min(indices.size(),initial.size()) ),
	m_indices_column(*this, boost::mem_fn(&detail::sample::indices) ),
	m_cn2_column(*this, boost::mem_fn(&detail::sample::cn2) ),
	m_windspeed_column(*this, boost::mem_fn(&detail::sample::windspeed)) {
	
	using utils::algo::merge_join;

	merge_join( indices.begin(), indices.end(), initial.begin(), initial.end(),
		    boost::make_function_output_iterator( pack_input(*this) ),
		    boost::mem_fn( &atmos::indices::multiexp::storage::value_type::timestamp),
		    boost::mem_fn( &atmos::variableexposure::initial::storage::value_type::timestamp) );
}
} // input 
namespace profile {
	
namespace detail {
	template<class T,class F> 
	boost::tuple< trm::stop::basic<T>, typename trm::type_traits<T>::function_type, typename trm::type_traits<T>::vector_type >
	steihaug_toint_regularization( const ublas::vector<T>& x0, F f, const trm::stop::basic<T>& stop_strategy, T reg_alpha, const ublas::vector<T>& reg_psi ) {
		trm::basic_driver< T, trm::delta::initial::nearest_cauchy,
			trm::core::steihaug_toint_nonnegative,
			trm::delta::update::simple_update,
			trm::stop::basic<T> > dr(x0,trm::functor::regularization<T>(trm::functor::least_squares<T>(f),trm::functor::vector_norm<T>(reg_alpha,reg_psi)),stop_strategy);
		return dr.minimize();
	}
}

storage::storage( size_type reserve ):
	m_residual_norm_column(*this, boost::mem_fn(&detail::sample::residual_norm) ),
	m_residual_column(*this, boost::mem_fn(&detail::sample::residual) ),
	m_cn2_column(*this, boost::mem_fn(&detail::sample::cn2) ),
	m_cn2_weight_column(*this, boost::mem_fn(&detail::sample::cn2_weight) ),
	m_windspeed_column(*this, boost::mem_fn(&detail::sample::windspeed) ),
	m_windspeed_weight_column(*this, boost::mem_fn(&detail::sample::windspeed_weight)) {
}

const double functor::cn2_scale = 1e-14;
const double functor::wind_scale = 100;

boost::tuple< vector_type, matrix_type, symmetric_matrix_type > functor::objective::operator() (const vector_type& x) const {
	typedef matrix_type::size_type size_type;
	
	const size_type profile_size = context().m_grid.size();
	vector_type c(indices_.size());
	matrix_type jacobian( indices_.size(), x.size() );
	symmetric_matrix_type hesse = zero_matrix<value_type>(x.size(),x.size());
	
	matrix_type mixed_derivative(indices_.size(),profile_size);
	matrix_type second_derivative(indices_.size(),profile_size);
	
	vector_range<const vector_type> cn2(x, range(0, profile_size));
	vector_range<const vector_type> windspeed(x, range(profile_size, x.size()));
        
	value_type tau = 1e-3;
	for( size_type i = 0; i < indices_.size() / atmos::indices_size; ++i,tau *= 2 ) {
		for( size_type j = 0; j < profile_size; ++j ) {
			boost::tuple<vector_type,vector_type,vector_type> weight = context().eval_weight(j,tau*windspeed(j)*context_type::wind_scale);

			// FIXME: check slice(j,0,atmos::indices_size)
			matrix_vector_slice<matrix_type> cn2_cell(jacobian, slice(i*atmos::indices_size,1,atmos::indices_size), slice(j,0,atmos::indices_size));
			matrix_vector_slice<matrix_type> windspeed_cell(jacobian, slice(i*atmos::indices_size,1,atmos::indices_size), slice(j+profile_size,0,atmos::indices_size));
			matrix_vector_slice<matrix_type> mixed_cell(mixed_derivative, slice(i*atmos::indices_size,1,atmos::indices_size), slice(j,0,atmos::indices_size));
			matrix_vector_slice<matrix_type> second_cell(second_derivative, slice(i*atmos::indices_size,1,atmos::indices_size), slice(j,0,atmos::indices_size));
			
			cn2_cell       = context_type::cn2_scale*weight.get<0>();
			windspeed_cell = context_type::cn2_scale*cn2(j)*tau*context_type::wind_scale*weight.get<1>();
			mixed_cell     = context_type::cn2_scale*tau*context_type::wind_scale*weight.get<1>();
			second_cell    = context_type::cn2_scale*cn2(j)*std::pow(tau*context_type::wind_scale,2.0)*weight.get<2>();
		}
		for( size_type k = 0; k < atmos::indices_size; ++k ) {
			c(i*atmos::indices_size+k) = inner_prod(subrange(row(jacobian,i*atmos::indices_size+k),0,profile_size),cn2)-indices_(i*atmos::indices_size+k);
		}
	}
	
	
	c = prod(context().m_weight_matrix,c);  // c <- Wc
	jacobian = prod(context().m_weight_matrix,jacobian);  // J <- WJ
	// H <- <WW^Tc|H>
	vector_type dc = prod(trans(context().m_weight_matrix),c);
	for( size_type i = 0; i < indices_.size(); ++i ) {
		for( size_type kj = 0; kj < profile_size; ++kj ) {
			hesse(kj,kj+profile_size) += dc(i) * mixed_derivative(i,kj);
			hesse(kj+profile_size,kj+profile_size) +=  dc(i) * second_derivative(i,kj);
		}
	}

	return boost::make_tuple(c,jacobian,hesse);
}

double functor::smooth_limit( double height ) {
	return 2 * std::sqrt( height * 1000 * 5e-6 ) * 100;
}
boost::tuple<vector_type,vector_type,vector_type> functor::eval_weight( grid_type::size_type height, double smooth ) const {
	assert( height < m_variableexposure_functions.size() );

	return  m_variableexposure_functions[height](std::abs(smooth));
}
functor::functor( const grid_type& grid, const indices_covariance_type& cov_matrix, double mz, const weight_type& weight ):
	m_grid(grid*mz), m_mz(mz) {
	using namespace utils::ublas;

	boost::tie( m_weight_matrix, m_invweight_matrix ) = utils::ublas::gauss_weight_matrix(cov_matrix);
	
	m_variableexposure_functions.reserve( m_grid.size() );
	for( grid_type::size_type i=0; i<m_grid.size(); ++i ) {
		m_variableexposure_functions.push_back( weight(m_grid(i)) );
	}
}

functor::result_type functor::operator() ( const indices_type& indices, const cn2_initial_type& cn2, const windspeed_initial_type& windspeed ) const {
	objective obj_function(indices,*this);
	vector_type initial_x(cn2.size()+windspeed.size());
	subrange(initial_x, 0, cn2.size()) = cn2/functor::cn2_scale;
	subrange(initial_x, cn2.size(), initial_x.size()) = (windspeed*100.0/*cm*/)/functor::wind_scale;
	
	value_type residual_norm;
	vector_type residual;
	vector_type ret;
	trm::stop::basic<double> stop_strategy;
	const double eps = std::numeric_limits<double>::epsilon();
	//boost::tie( stop_strategy, residual_norm, ret) = trm::steihaug_toint_nonnegative_least_squares(initial_x,obj_function,trm::stop::basic<double>(300,50000*eps,10*eps,10*eps));
	boost::tie( stop_strategy, residual_norm, ret) = detail::steihaug_toint_regularization(initial_x,obj_function,trm::stop::basic<double>(300,50000*eps,10*eps,10*eps),1e-2,initial_x);
	vector_type ret_weight(ret.size());
	matrix_type jacobian;
	symmetric_matrix_type hesse;
	boost::tie( residual, jacobian, hesse ) = obj_function(ret);
	
	for( size_type i=0; i < cn2.size(); ++i ) {
		ret_weight(i) = 1.0;
		if( ret(i) == 0.0 ) {
			ret_weight(i+cn2.size()) = 0.0;
		} else {
			ret_weight(i+cn2.size()) = 1.0;
		}
	}
	
	return boost::make_tuple( /*stop_strategy.iternum() > 0 */stop_strategy.grad_flag(),
				  residual_norm, residual,
				  subrange(ret, 0, cn2.size())*functor::cn2_scale,
				  subrange(ret_weight, 0, cn2.size())*functor::cn2_scale,
				  subrange(ret, cn2.size(), ret.size())*functor::wind_scale/100.0/*cm*/,
				  subrange(ret_weight, cn2.size(), ret_weight.size())*functor::wind_scale/100.0/*cm*/ );
}

double functor::seeing( const cn2_type& cn2 ) const {
	const double JPOW_SEE = 0.6 ;
	const double KL = 1.1486983549970351; // 0.5 ** LPOW_SEE
	const double KS = 1.73588e+07;
	return KS*KL*std::pow( sum( cn2 ), JPOW_SEE );
}

double functor::tau0( const cn2_type& cn2, const windspeed_type& wind ) const {
	using namespace utils::ublas;
	
	const double WPOW_TAU = -0.6;
	const double KL = 0.43527528164806206; // 0.5 ** 6/5;
	const double KS = 0.057 * 6.3095734448019429e-08 * 1e03;
	return KS*KL*pow( sum( element_prod( cn2, pow( wind, -1.0/WPOW_TAU ) ) ), WPOW_TAU );
}

residual_type functor::indices_residual( const residual_type& residual ) const {
	return prod( m_invweight_matrix, residual );
}

} // profile
}}
