/*
    $Id: variableexposure_profile.h 244 2011-08-28 08:14:30Z matwey $
    Copyright (C) 2011  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _VARIATEEXPOSURE_PROFILE_H
#define _VARIATEEXPOSURE_PROFILE_H

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/function.hpp>

#include "indices.h"
#include "profile.h"
#include "shortexposure_profile.h"
#include "weif.h"
#include "worksheet.h" // local_context

namespace atmos {
namespace variableexposure {
namespace initial {
	using namespace boost::numeric::ublas;
	
	typedef boost::posix_time::ptime time_type;
	typedef atmos::profile::cn2_type cn2_type;
	typedef atmos::shortexposure::profile::windspeed_type windspeed_type;
	
namespace detail {
	struct sample {
		typedef atmos::variableexposure::initial::time_type      timestamp_type;
		typedef atmos::variableexposure::initial::cn2_type       cn2_type;
		typedef atmos::variableexposure::initial::windspeed_type windspeed_type;

		timestamp_type timestamp;
		cn2_type       cn2;
		windspeed_type windspeed;

		sample( const timestamp_type& timestamp, const cn2_type& cn2, const windspeed_type& windspeed ):
			timestamp(timestamp), cn2(cn2), windspeed(windspeed) {
		}
	};
} // detail

class storage:
	public table_storage< detail::sample > {
public:
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::cn2_type, detail::sample > >       cn2_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::windspeed_type, detail::sample > > windspeed_column_type;
private:
	cn2_column_type       m_cn2_column;
	windspeed_column_type m_windspeed_column;
public:
	storage( size_type reserve = 60 );
	storage( const atmos::profile::storage& profile, const atmos::shortexposure::profile::storage& windspeed );

	const cn2_column_type&       cn2()       const { return m_cn2_column; }
	const windspeed_column_type& windspeed() const { return m_windspeed_column; }

	void push_back( const time_type& timestamp, const cn2_type& cn2, const windspeed_type& windspeed ) {
		table_storage< detail::sample >::push_back( detail::sample(timestamp, cn2, windspeed) );
	}
	
	void repair_initial_cn2( const cn2_type& avg_cn2 );
	void repair_initial_windspeed( const windspeed_type& avg_windspeed );
};
} // initial

namespace input {
	using namespace boost::numeric::ublas;

	typedef boost::posix_time::ptime time_type;
	typedef atmos::indices::multiexp::indices_vector_type indices_vector_type;
	typedef atmos::indices::vector_type         vector_type;
	typedef symmetric_matrix<vector_type::value_type> symmetric_matrix_type;
	typedef banded_matrix<vector_type::value_type> banded_matrix_type;
	
	typedef vector_type indices_type;
	typedef banded_matrix_type indices_covariance_type;
	//typedef symmetric_matrix_type indices_covariance_type;
	
	typedef atmos::variableexposure::initial::cn2_type cn2_type;
	typedef atmos::variableexposure::initial::windspeed_type windspeed_type;

namespace detail {
	struct sample {
		typedef atmos::variableexposure::input::time_type timestamp_type;
		typedef atmos::variableexposure::input::indices_type indices_type;
		typedef atmos::variableexposure::input::cn2_type cn2_type;
		typedef atmos::variableexposure::input::windspeed_type windspeed_type;

		timestamp_type timestamp;
		indices_type   indices;
		cn2_type       cn2;
		windspeed_type windspeed;

		sample( const timestamp_type& timestamp, const indices_type& indices, const cn2_type& cn2, const windspeed_type& windspeed ):
			timestamp(timestamp), indices(indices), cn2(cn2), windspeed(windspeed) {
		}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::indices_type, detail::sample > > indices_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::cn2_type, detail::sample > > cn2_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::windspeed_type, detail::sample > > windspeed_column_type;
private:
	indices_column_type m_indices_column;
	cn2_column_type m_cn2_column;
	windspeed_column_type m_windspeed_column;
public:
	storage( size_type reserve = 60 );
	storage( const atmos::indices::multiexp::storage& indices, const atmos::variableexposure::initial::storage& initial );

	const indices_column_type&   indices()   const { return m_indices_column; }
	const cn2_column_type&       cn2()       const { return m_cn2_column; }
	const windspeed_column_type& windspeed() const { return m_windspeed_column; }

	void push_back( const time_type& timestamp, const indices_type& indices, const cn2_type& cn2, const windspeed_type& windspeed ) {
		table_storage< detail::sample >::push_back( detail::sample( timestamp, indices, cn2, windspeed ) );
	}
};
} // input

namespace profile {
	using namespace boost::numeric::ublas;
	
	typedef boost::posix_time::ptime time_type;
	typedef double                 value_type;
	typedef vector<value_type>     vector_type;
	typedef matrix<value_type>     matrix_type;
	typedef banded_matrix<value_type> banded_matrix_type;
	typedef symmetric_matrix<value_type> symmetric_matrix_type;
	typedef vector<vector_type>    vector_vector_type;

	typedef vector_type cn2_type;
	typedef vector_type windspeed_type;
	typedef value_type  residual_norm_type;
	typedef vector_type residual_type;
	
namespace detail {
	struct sample {
		typedef atmos::variableexposure::profile::cn2_type       cn2_type;
		typedef atmos::variableexposure::profile::windspeed_type windspeed_type;
		typedef atmos::variableexposure::profile::residual_norm_type residual_norm_type;
		typedef atmos::variableexposure::profile::residual_type residual_type;

		residual_norm_type residual_norm;
		residual_type  residual;
		cn2_type       cn2;
		cn2_type       cn2_weight;
		windspeed_type windspeed;
		windspeed_type windspeed_weight;

		sample( const residual_norm_type& residual_norm, const residual_type& residual, const cn2_type& cn2, const cn2_type& cn2_weight, const windspeed_type& windspeed, const windspeed_type& windspeed_weight ):
			residual_norm(residual_norm), residual(residual), cn2(cn2), cn2_weight(cn2_weight), windspeed(windspeed), windspeed_weight(windspeed_weight) {
		}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::residual_norm_type, detail::sample > > residual_norm_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::residual_type, detail::sample > >  residual_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::cn2_type, detail::sample > >       cn2_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::cn2_type, detail::sample > >       cn2_weight_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::windspeed_type, detail::sample > > windspeed_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::windspeed_type, detail::sample > > windspeed_weight_column_type;
private:
	residual_norm_column_type    m_residual_norm_column;
	residual_column_type  m_residual_column;
	cn2_column_type       m_cn2_column;
	cn2_weight_column_type m_cn2_weight_column;
	windspeed_column_type m_windspeed_column;
	windspeed_weight_column_type m_windspeed_weight_column;
public:
	storage( size_type reserve = 60 );

	const residual_norm_column_type& residual_norm() const { return m_residual_norm_column; }
	const residual_column_type&  residual()  const { return m_residual_column; }
	const cn2_column_type&       cn2()       const { return m_cn2_column; }
	const cn2_weight_column_type& cn2_weight() const { return m_cn2_weight_column; }
	const windspeed_column_type& windspeed() const { return m_windspeed_column; }
	const windspeed_weight_column_type& windspeed_weight() const { return m_windspeed_weight_column; }

	void push_back( const residual_norm_type& residual_norm, const residual_type& residual, const cn2_type& cn2, const cn2_type& cn2_weight, const windspeed_type& windspeed, const windspeed_type& windspeed_weight ) {
		table_storage< detail::sample >::push_back( detail::sample(residual_norm, residual, cn2, cn2_weight, windspeed, windspeed_weight) );
	}
};

class functor {
	friend class local_context<const functor>;
	public:
		typedef boost::tuple<bool,residual_norm_type,residual_type,cn2_type,cn2_type,windspeed_type,windspeed_type> result_type;
		typedef atmos::variableexposure::input::indices_type indices_type;
		typedef atmos::variableexposure::input::indices_covariance_type indices_covariance_type;
		typedef atmos::variableexposure::input::cn2_type       cn2_initial_type;
		typedef atmos::variableexposure::input::windspeed_type windspeed_initial_type;

		typedef vector_type grid_type;
		typedef atmos::weight::factory::variableexposure_factory_type::value_type weight_type;
		
		class objective: public local_context<const functor> {
		public:
			objective( const indices_type& indices, context_reference_type ctx ): indices_(indices), local_context_type( ctx ) {
			}
			boost::tuple< vector_type, matrix_type, symmetric_matrix_type > operator() (const vector_type& x) const;
		private:
			const indices_type& indices_;
		};
		
		static const double cn2_scale;
		static const double wind_scale;
	private:
		typedef weight_type::result_type weight1_type;
		grid_type     m_grid;
		double        m_mz;
		matrix_type   m_weight_matrix;
		matrix_type   m_invweight_matrix;
		
		std::vector< weight1_type > m_variableexposure_functions;

		static double smooth_limit( double height );
		boost::tuple<vector_type,vector_type,vector_type> eval_weight( grid_type::size_type height, double smooth ) const;
	public:
		functor( const grid_type& grid, const indices_covariance_type& cov_matrix, double mz, const weight_type& weight );

		result_type operator() ( const indices_type& indices, const cn2_initial_type& cn2, const windspeed_initial_type& windspeed ) const;
		
		double seeing( const cn2_type& cn2 ) const;
		double tau0( const cn2_type& cn2, const windspeed_type& wind ) const;
		residual_type indices_residual( const residual_type& residual ) const;
};
} // profile

}}

#endif // _VARIATEEXPOSURE_PROFILE_H
