/*
    $Id: weif.cpp 235 2011-07-17 16:10:52Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <boost/function.hpp>

#include <boost/numeric/ublas/io.hpp>
#include <boost/math/constants/constants.hpp>

#include "atmos.h"
#include "weif.h"

#include "gsl_wrap.h"
#include "fft.h"

namespace atmos {
namespace weight {

const double POWIND = -2.66666666667;
const double POWINDLONG = -3.66666666667;
const double POWDIFF = -0.666666666667;
const double MKM2CM = 1E-4;
const double EPS = 1.0e-4 ;
const double CALIBR  = 44.62621363;
const double DIMMCONST = 2.403*4.64159;
const double KM2CM  = 1E5 ;
const int  WORKSPACE = 10000;

value_type wind_factor(const value_type& xi) {
	using namespace boost::math::constants;
	
	/* temporal rational approximation */
	if( xi == value_type(0) ) return value_type(1);
	if( xi < 0.3 ) return 1.0 - pi<value_type>()*pi<value_type>()*xi*xi/6;
	if( xi < 2.0 ) 
		return (-0.8031909e-1*xi*xi*xi+0.38761586*xi*xi-0.48643327*xi+0.25456108)/(0.4178926e-1*xi*xi*xi+0.3605432*xi*xi-0.41300218*xi+0.254561095);
	if( xi < 5.0 )
		return (0.4e-5*xi*xi*xi+0.14804*xi*xi+0.1627*xi-0.13e-2)/(0.4670*xi*xi*xi+0.488*xi*xi+0.3e-1*xi-0.4e-2);
	/*if( xi > 5.0 )*/
	return 1.0/(xi*pi<value_type>());
	
	/* Good implementation should be
	 * like this:

	using utils::math::struve_h;
	using boost::math::cyl_bessel_j;
	using namespace boost::math::constants;
	
	const value_type x = 2*xi*pi<value_type>();
	const value_type b0 = cyl_bessel_j(0,x);
	const value_type b1 = cyl_bessel_j(1,x);

	return 2*b0-2*b1/x-pi<value_type>()*(b0*struve_h(1,x)-b1*struve_h(0,x));
	*/
}

static gsl::set_error_handler_off set_error_handler_off_;

struct mass_params {
	typedef boost::function1< std::complex<double>, double > sfun_type;

	double diameter1;
	double diameter2;
	double diameter3;
	double diameter4;
	double PIx2lambda0;
	sfun_type sfun;
	double altitude;
	double powind;
	double smooth;

	mass_params( const sfun_type& sfun ): sfun(sfun), smooth(0.0) {}
};

struct dimm_params {
	typedef boost::function1< std::complex<double>, double > sfun_type;

	double aperture;
	double dimmbase;
	double PIx2lambda0;
	double phi;
	sfun_type sfun;
	double altitude;

	dimm_params( const sfun_type& sfun ): sfun(sfun) {}
};

double mass_integrand(double x, void * params) {
	using namespace std;
	
	mass_params p = *( mass_params *) params;
	double p1 = 0;
	double p3 = 0;
	double f = p.altitude*x*x/2;
	const std::complex<double>& sfun_f = p.sfun(f);
	double sf = imag( sfun_f ) * cos( p.PIx2lambda0*f ) + real( sfun_f ) * sin( p.PIx2lambda0*f );
	sf = sf*sf*pow( x, p.powind );
	if( p.smooth > 0 ) {
		sf *= wind_factor( x*p.smooth );
	}
	if( p.diameter1 != p.diameter2 ) {
		p1 = ( p.diameter1*j1( p.diameter1*x ) - p.diameter2*j1( p.diameter2*x ) ) / ( p.diameter1*p.diameter1 - p.diameter2*p.diameter2 );
	}
	if( p.diameter3 != p.diameter4 ) {
		p3 = ( p.diameter3*j1( p.diameter3*x ) - p.diameter4*j1( p.diameter4*x ) ) / ( p.diameter3*p.diameter3 - p.diameter4*p.diameter4 );
	}
	return sf * 4.0 * (p1 * p3) / (x*x);
}

double dimm_integrand(double x, void * params) {
	dimm_params p = *( dimm_params *) params;
	double f = p.altitude*x*x/2;
	double a = p.aperture*x;
	double b = p.dimmbase*x;
	const std::complex<double>& sfun_f = p.sfun(f);
	double cf = real( sfun_f ) * cos( p.PIx2lambda0*f ) - imag( sfun_f ) * sin( p.PIx2lambda0*f );
	cf = cf*8*jn(2,a)/a/a;
	cf = cf*cf*pow( x, POWDIFF );
	return cf*( 1.0 - 2.0*cos(p.phi)*cos(p.phi)*j0(b) + 2.0*cos(2*p.phi)*j1(b)/b );
}

size_t next_pow2( size_t x ) {
	if( !(x&(x-1)) ) return x;
	for( size_t i = 1; i < std::numeric_limits<size_t>::digits; i<<=1 ) {
		x |= (x >> i);
	}
	return (x & ~(x >> 1)) << 1;
}

template<class V,class V2> typename V::const_iterator lambda_effective( const V& lambda, const V2& dist ) {
	typedef typename V::value_type value_type;
	typedef typename V::const_iterator const_iterator;

	value_type leff = inner_prod( lambda, dist ) / sum( dist );
	const_iterator it = std::lower_bound( lambda.begin(), lambda.end(), leff );

	return it;
}

std::pair< vector_type, vector< complex_type > > spectral_band_fourier( const vector<double>& lambda, const vector<double>& dist, unsigned int center ) {
	typedef vector<double>::size_type size_type;
	
	const size_type grid_factor = 4;
	const double dlambda = lambda(1) - lambda(0);
	const size_type distft_size = 2 * grid_factor * next_pow2( dist.size() );
	
	vector< std::complex<double> > dist_ft = zero_vector< std::complex<double> >( distft_size );
	for( vector<double>::size_type l = 0; l < dist.size(); l++ ) {
		vector<double>::size_type j = ( l < center ) ? distft_size + l - center : l - center;
		dist_ft(j) = dist(l);
	}
	noalias( dist_ft ) = utils::ublas::fft_radix2( dist_ft );
	
	vector< double > freq( distft_size/2 );
	for( vector<double>::size_type i = 0; i < freq.size(); ++i ) {
		freq(i) = i/(dlambda*MKM2CM)/(distft_size);
	}
	
	return std::make_pair( freq, subrange( dist_ft, 0, distft_size/2 ) );
}

namespace detail {

basic_weight::value_type basic_weight::do_calculate( value_type height, value_type eps1, value_type eps2, value_type eps3, value_type eps4 ) const {
	gsl::integration_workspace w(WORKSPACE);
	double error;
	mass_params params( mass_sfourier_ );
	gsl_function F;
	F.function = &mass_integrand;
	F.params = &params;
	params.diameter1 = M_PI*eps1;
	params.diameter2 = M_PI*eps2;
	params.diameter3 = M_PI*eps3;
	params.diameter4 = M_PI*eps4;
	params.PIx2lambda0 = 2*M_PI * mass_leff_ * MKM2CM;
	params.powind = POWIND + freq_power_;
	double result;
	params.altitude = height*KM2CM;
	double hi_limit = std::sqrt( 2 * mass_freq_max_ / params.altitude );
	if( std::isfinite( hi_limit ) )
		gsl::integration_qags( &F, 0, hi_limit, 0, EPS, WORKSPACE, w, &result, &error);
	else
		gsl::integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result, &error);
	return result*CALIBR*std::pow(M_PI*100,freq_power_);
}

basic_weight::basic_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& freq_power ):
	ap_inn_( ap_inn ),
	ap_out_( ap_out ),
	mass_leff_index_( lambda_effective( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ) ).index() ),
	mass_leff_( mass_lambda( mass_leff_index_ ) ),
	mass_sfourier_( spectral_band_fourier( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ), mass_leff_index_ ) ),
	freq_power_(freq_power) {

	mass_freq_max_ = mass_sfourier_.first_data()( mass_sfourier_.size() - 1 );
}

basic_weight::result_type basic_weight::operator() ( const arg_type& height ) const {
	const size_type mass_nchan = atmos::channels;
	vector_type ret( atmos::indices_size );
	size_type cur_n = 0;

	for( size_type i = 0; i < mass_nchan; ++i )
		ret( cur_n++ ) = do_calculate( height, ap_out_(i), ap_inn_(i), ap_out_(i), ap_inn_(i) );

	for( size_type i = 0; i < mass_nchan; ++i )
		for( size_type j = i + 1; j < mass_nchan; j++ )
			ret( cur_n++ ) = do_calculate( height, ap_out_(i), ap_inn_(i), ap_out_(j), ap_inn_(j) );

	return ret;
}

}

massdimm_weight::value_type massdimm_weight::do_dimm_calculate( value_type height, value_type diam, value_type dbase, value_type phi ) const {
	gsl::integration_workspace w(WORKSPACE);
	double error;
	dimm_params params(dimm_sfourier_);
	gsl_function F;
	F.function = &dimm_integrand;
	F.params = &params;
	params.aperture = M_PI*diam;
	params.dimmbase = 2*M_PI*dbase;
	params.PIx2lambda0 = 2*M_PI*dimm_leff_*MKM2CM;
	params.phi = phi;
	double result;
	params.altitude = height*KM2CM;
	double hi_limit = std::sqrt( 2 * dimm_freq_max_ / params.altitude );
	if( std::isfinite( hi_limit ) )
		gsl::integration_qags( &F, 0, hi_limit, 0, EPS, WORKSPACE, w, &result, &error);
	else
		gsl::integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result, &error);
	return result*DIMMCONST;
}

massdimm_weight::massdimm_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& dimm_lambda, const vector_type& dimm_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& dimmsize, const value_type& dimmbase ):
	basic_weight( mass_lambda, mass_response, ap_inn, ap_out ),
	dimmsize_( dimmsize * 100 ), dimmbase_( dimmbase * 100 ),
	dimm_leff_index_( lambda_effective( dimm_lambda, dimm_response ).index() ),
	dimm_leff_( dimm_lambda( dimm_leff_index_ ) ),
	dimm_sfourier_( spectral_band_fourier( dimm_lambda, dimm_response, dimm_leff_index_ ) ) {

	dimm_freq_max_ = dimm_sfourier_.first_data()( dimm_sfourier_.size() - 1 );
}

massdimm_weight::result_type massdimm_weight::operator() ( const arg_type& height ) const {
	vector_type ret( atmos::indices_size + atmos::dimm_indices_size );
	size_type cur_n = atmos::indices_size;

	subrange( ret, 0, atmos::indices_size ) = basic_weight::operator() ( height );
	
	ret( cur_n++ ) = do_dimm_calculate( height, dimmsize_, dimmbase_, 0.0 ) * 1e+11;
	ret( cur_n++ ) = do_dimm_calculate( height, dimmsize_, dimmbase_, M_PI/2 ) * 1e+11;

	return ret;
}

mass_longexposure_weight::mass_longexposure_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out ):
	basic_weight( mass_lambda, mass_response, ap_inn, ap_out, -1.0 ) {
}

mass_shortexposure_weight::mass_shortexposure_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out ):
	basic_weight( mass_lambda, mass_response, ap_inn, ap_out, 2.0 ) {
}

mass_variableexposure_weight::mass_variableexposure_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out ):
	ap_inn_( ap_inn ),
	ap_out_( ap_out ),
	mass_leff_index_( lambda_effective( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ) ).index() ),
	mass_leff_( mass_lambda( mass_leff_index_ ) ),
	mass_sfourier_( spectral_band_fourier( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ), mass_leff_index_ ) ) {

	mass_freq_max_ = mass_sfourier_.first_data()( mass_sfourier_.size() - 1 );
}
mass_variableexposure_weight::result_type mass_variableexposure_weight::operator() ( const arg_type& height, const arg_type& smooth ) const {
	const size_type mass_nchan = atmos::channels;
	vector_type ret( atmos::indices_size );
	size_type cur_n = 0;

	for( size_type i = 0; i < mass_nchan; ++i )
		ret( cur_n++ ) = do_calculate( height, smooth, ap_out_(i), ap_inn_(i), ap_out_(i), ap_inn_(i) );

	for( size_type i = 0; i < mass_nchan; ++i )
		for( size_type j = i + 1; j < mass_nchan; j++ )
			ret( cur_n++ ) = do_calculate( height, smooth, ap_out_(i), ap_inn_(i), ap_out_(j), ap_inn_(j) );

	return ret;
}
mass_variableexposure_weight::value_type mass_variableexposure_weight::do_calculate( const value_type& height, const value_type& smooth, const value_type& eps1, const value_type& eps2, const value_type& eps3, const value_type& eps4 ) const {
	gsl::integration_workspace w(WORKSPACE);
	double error;
	mass_params params( mass_sfourier_ );
	gsl_function F;
	F.function = &mass_integrand;
	F.params = &params;
	params.diameter1 = M_PI*eps1;
	params.diameter2 = M_PI*eps2;
	params.diameter3 = M_PI*eps3;
	params.diameter4 = M_PI*eps4;
	params.PIx2lambda0 = 2*M_PI * mass_leff_ * MKM2CM;
	params.powind = POWIND;
	params.smooth = smooth;
	double result;
	params.altitude = height*KM2CM;
	double hi_limit = std::sqrt( 2 * mass_freq_max_ / params.altitude );
	if( std::isfinite( hi_limit ) )
		gsl::integration_qags( &F, 0, hi_limit, 0, EPS, WORKSPACE, w, &result, &error);
	else
		gsl::integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result, &error);
	return result*CALIBR;
}

}

linear_generator::linear_generator( result_type start, result_type step ): start_(start), step_(step) {
}
linear_generator::result_type linear_generator::operator()( size_type i ) const {
	return ( start_ + step_ * i );
}
linear_generator::size_type linear_generator::size( result_type max ) const {
	return static_cast<size_type>( std::floor( (max - start_) / step_ ) );
}

lg_generator::lg_generator( result_type start, result_type lstep, result_type gstep ): start_(start), t_(lstep), q_(result_type(1) + gstep / 2) {
	max_ = static_cast< size_type >( std::floor( result_type(2) / gstep - start / lstep ) );
}
lg_generator::result_type lg_generator::operator()( size_type i ) const {
	if( i <= max_ ) return ( start_ + t_ * i );
	return ( start_ + t_ * max_ ) * std::pow( q_, static_cast<int>(i - max_) ); // FIXME: pow(double,int) for <4.2.1
}
lg_generator::size_type lg_generator::size( result_type max ) const {
	result_type gmax = std::floor( std::log( max / ( start_ + t_ * max_ ) ) / std::log( q_ ) );
	if( gmax > result_type(0) ) return static_cast< size_type >( gmax ) + max_ + 1;
	return static_cast< size_type >(std::floor( (max - start_) / t_ ));
}

}