/*
    $Id: weight_factory.cpp 235 2011-07-17 16:10:52Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <iomanip>
#include <fstream>
#include <cmath>

#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem/operations.hpp>

#include "atmos.h"
#include "io.h"
#include "weight_factory.h"
#include "weif.h"
#include "utils.h"

namespace atmos {
namespace weight {

namespace detail {
weight_wrapper::weight_wrapper( context_reference_type ctx, const weight_type& weight ): local_context_type(ctx), weight_(weight) {
}

weight_wrapper::result_type weight_wrapper::operator() (const value_type& x) const {
	if( x < context().m_basis_grid(0) || x > context().m_basis_grid( context().m_basis_grid.size()-1 ) ) {
		throw std::out_of_range( (boost::format("Grid height %1% is out of ragne") % x ).str() );
	}
	return weight_(x);
}

mass_weight_key::mass_weight_key( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam ):
	spectype(spectype), inner_diam(inner_diam), outer_diam(outer_diam) {
}


variableexposure_weight_wrapper::variableexposure_weight_wrapper( context_reference_type ctx, const weight_type& weight ): local_context_type(ctx) {
	typedef cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > spline_type;

	weight_.reserve(context().m_smooth_grid.size());
	
	for( size_t i = 0; i < context().m_smooth_grid.size(); ++i ) {
		vector_vector_type tmp( context().m_basis_grid.size() );
		for( vector_type::size_type j = 0; j < tmp.size(); ++j ) {
			const value_type smooth_limit = 2 * std::sqrt( context().m_basis_grid(j) * 1000 * 5e-6 ) * 100;
			tmp(j) = weight( context().m_basis_grid(j), context().m_smooth_grid(i)*smooth_limit );
		}
		weight_.push_back( spline_type(context().m_basis_grid, tmp) );
	}
}

struct spline_and_asymp {
	typedef cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > spline_type;
	spline_and_asymp( const spline_type& spline ):
		spline_(spline),
		last_node_(spline.first_data()( spline.first_data().size()-1 )),
		last_value_(spline.second_data()( spline.second_data().size()-1 )) { }

	boost::tuple<vector_type,vector_type,vector_type> operator() (const value_type& smooth) const {
		if( smooth < spline_.first_data()(0) ) {
			throw std::out_of_range( (boost::format("Smooth %1% is out of range") % smooth ).str() );
		}	
		if( smooth >= last_node_ ) {
			return boost::make_tuple( last_value_*last_node_/smooth,
						  last_value_*last_node_/(-smooth*smooth),
						  last_value_*2.0*last_node_/(smooth*smooth*smooth) );
		}
		return spline_(smooth,true);
	}
private:
	spline_type spline_;
	value_type last_node_;
	vector_type last_value_;
};

variableexposure_weight_wrapper::result_type variableexposure_weight_wrapper::operator() (const value_type& x) const {
	typedef cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > spline_type;

	if( x < context().m_basis_grid(0) || x > context().m_basis_grid( context().m_basis_grid.size()-1 ) ) {
		throw std::out_of_range( (boost::format("Grid height %1% is out of ragne") % x ).str() );
	}
	
	const value_type smooth_limit = std::max( 1.0, 2 * std::sqrt( x * 1000 * 5e-6 ) * 100 );
	
	vector_vector_type tmp(context().m_smooth_grid.size());
	for( vector_type::size_type i = 0; i < context().m_smooth_grid.size(); ++i ) {
		tmp(i) = weight_[i](x);
	}
	
	return spline_and_asymp( spline_type( context().m_smooth_grid * smooth_limit, tmp,
		vector_type(zero_vector<value_type>(tmp(0).size())),
		vector_type(-tmp(tmp.size()-1)/(smooth_limit*context().m_smooth_grid(context().m_smooth_grid.size()-1))) ) );
}

bool mass_weight_key::operator==( const mass_weight_key& key ) const {
	return !(*this < key) && !(key < *this);
}
bool mass_weight_key::operator!=( const mass_weight_key& key ) const {
	return *this < key || key < *this;
}
bool mass_weight_key::operator< ( const mass_weight_key& key ) const {
	if( nocase_less_( spectype, key.spectype ) ) {
		return true;
	} else if( nocase_less_( key.spectype, spectype ) ) {
		return false;
	}
	if( std::lexicographical_compare( inner_diam.begin(), inner_diam.end(), key.inner_diam.begin(), key.inner_diam.end() ) ) {
		return true;
	} else if( std::lexicographical_compare( key.inner_diam.begin(), key.inner_diam.end(), inner_diam.begin(), inner_diam.end() ) ) {
		return false;
	}
	if( std::lexicographical_compare( outer_diam.begin(), outer_diam.end(), key.outer_diam.begin(), key.outer_diam.end() ) ) {
		return true;
	} else if( std::lexicographical_compare( key.outer_diam.begin(), key.outer_diam.end(),  outer_diam.begin(), outer_diam.end() ) ) {
		return false;
	}
	
	return false;
}
std::ostream& mass_weight_key::print( std::ostream& stm ) const {
	return stm << spectype;
}

massdimm_weight_key::massdimm_weight_key( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam, double dimmbase, double dimmsize ):
	mass_weight_key(spectype, inner_diam, outer_diam),
	dimmbase(dimmbase), dimmsize(dimmsize) {
}

bool massdimm_weight_key::operator==( const massdimm_weight_key& key ) const {
	return !(*this < key) && !(key < *this);
}
bool massdimm_weight_key::operator!=( const massdimm_weight_key& key ) const {
	return *this < key || key < *this;
}
bool massdimm_weight_key::operator< ( const massdimm_weight_key& key ) const {
	if( mass_weight_key::operator< (key) ) {
		return true;
	} else if( key.mass_weight_key::operator< (*this) ) {
		return false;
	}
	
	if( dimmbase < key.dimmbase ) { return true; } else if( key.dimmbase < dimmbase ) { return false; }
	if( dimmsize < key.dimmsize ) { return true; } else if( key.dimmsize < dimmsize ) { return false; }
	
	return false;
}
std::ostream& operator<< ( std::ostream& stm, const mass_weight_key& key ) {
	return key.print(stm);
}
std::ostream& operator<< ( std::ostream& stm, const massdimm_weight_key& key ) {
	return key.print(stm);
}
}

std::pair< vector_type, vector_type > load_reaction( const boost::filesystem::path& file ) {
	std::ifstream fstm( file.string().c_str() );
	if( fstm.fail() )
		throw std::runtime_error( "Can't open file " + file.string() );

	std::pair< vector_type, vector_type > ret;
	vector< double >::size_type alloc_size = 10;
	vector< double >::size_type size = 0;

	ret.first.resize( alloc_size );
	ret.second.resize( alloc_size );

	std::string cur_line;
	double cx,cy;
	while( std::getline( fstm, cur_line ) ) {
		boost::trim_if( cur_line, boost::is_space() );
		if( cur_line.size() < 1 || cur_line.at(0) == '#' ) continue;
		std::istringstream cstm( cur_line );
		if( !(cstm >> cx >> cy) )
			throw std::runtime_error( "File format error in " + file.string() );
		ret.first( size ) = cx;
		ret.second( size ) = cy;
		++size;
		if( size >= ret.first.size() ) {
			alloc_size += alloc_size;
			ret.first.resize( alloc_size );
			ret.second.resize( alloc_size );
		}
	}

	ret.first.resize( size );
	ret.second.resize( size );

	/* Post check */
	if( size == 0 )
		throw std::runtime_error( "File " + file.string() + " contains no data" );
	if( std::adjacent_find( ret.first.begin(), ret.first.end(), std::greater<double>() ) != ret.first.end() )
		throw std::runtime_error( "File " + file.string() + " contains unsorted data" );
	if( std::count_if( ret.first.begin(), ret.first.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "File " + file.string() + " contains negative wavelengths" );
	if( std::count_if( ret.second.begin(), ret.second.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "File " + file.string() + " contains negative data" );

	ret.first = ret.first * 0.001; // convert from nm to mkm

	return ret;
}

std::pair< vector_type, vector_type > photon_distribution( const vector_type& spectral_lambda, const vector_type& spectral_response, const vector_type& device_lambda, const vector_type& device_response ) {
	std::pair< vector_type, vector_type > ret;

	assert( spectral_lambda.size() == spectral_response.size() );
	assert( device_lambda.size() == device_response.size() );

	ret.first.resize( std::min(spectral_lambda.size(), device_lambda.size()) );

	vector_type::iterator it = std::set_intersection( spectral_lambda.begin(), spectral_lambda.end(), device_lambda.begin(), device_lambda.end(), ret.first.begin() );
	if( it - ret.first.begin() == 0 )
		throw std::runtime_error( "No match of response and spectrum grids" );

	ret.first.resize( it - ret.first.begin() );
	ret.second.resize( ret.first.size() );

	vector_type::const_iterator sit = spectral_lambda.begin();
	vector_type::const_iterator dit = device_lambda.begin();
	vector_type::iterator rit = ret.second.begin();
	for( vector_type::const_iterator it = ret.first.begin(); it != ret.first.end(); ++it ) {
		sit = std::find( sit, spectral_lambda.end(), *it );
		dit = std::find( dit, device_lambda.end(), *it );
		*rit = spectral_response( sit.index() ) * device_response( dit.index() );
		++rit;
	}

	if( std::count_if( ret.second.begin(), ret.second.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
		throw std::runtime_error( "SED contains negative data" );

	ret.second = ret.second / sum( ret.second );

	return ret;
}

void dump_weight( const boost::filesystem::path& file, const boost::function1< vector_type, double >& function, const vector_type& grid ) {
	static const char* name[] = { "A", "B", "C", "D", "AB", "AC", "AD", "BC", "BD", "CD", "L", "T" };
	static const unsigned int name_size = sizeof(name) / sizeof(name[0]);
	const double wscale = 1e+11;

	std::ofstream fstm( file.string().c_str() );
	if( fstm.fail() )
		throw std::runtime_error( "Can't open file " + file.string() );

	fstm.exceptions( std::ifstream::failbit | std::ifstream::badbit );

	fstm << "# Z " << std::setprecision(2) << grid << std::endl;
	fstm << "# W " << std::setprecision(0) << std::scientific << wscale << std::endl;
	if( grid.size() < 1 )
		 return;

	unsigned int result_size = function( grid(0) ).size();

	fstm << "## Alt";
	for( unsigned int i = 0; i < std::min(name_size, result_size); ++i )
		fstm << "\tW(" << name[i] << ")/" <<  std::scientific << wscale;
	fstm << std::endl;

	for( vector_type::size_type i = 0; i < grid.size(); ++i ){
		fstm << std::fixed << std::setprecision(3) << grid(i);
		vector_type result = function( grid(i) );
		fstm << std::scientific << std::setprecision(4);
		for( vector_type::size_type j = 0; j < result.size(); ++j ) {
			fstm << "\t" << result(j) / wscale;
		}
		fstm << std::endl;
	}
}

void dump_variable_weight( const boost::filesystem::path& file, const boost::function1< boost::function1 <boost::tuple<vector_type,vector_type,vector_type>, double >, double >& function,  const vector_type& grid ) {
	static const char* name[] = { "A", "B", "C", "D", "AB", "AC", "AD", "BC", "BD", "CD" };
	static const unsigned int name_size = sizeof(name) / sizeof(name[0]);
	const double wscale = 1e+11;

	std::ofstream fstm( file.string().c_str() );
	if( fstm.fail() )
		throw std::runtime_error( "Can't open file " + file.string() );

	fstm.exceptions( std::ifstream::failbit | std::ifstream::badbit );

	fstm << "# Z " << std::setprecision(2) << grid << std::endl;
	fstm << "# W " << std::setprecision(0) << std::scientific << wscale << std::endl;
	if( grid.size() < 1 )
		 return;

	const vector_type smooth_grid = vector_generator< linear_generator >( linear_generator(0.0, 0.5), 200 );
	for( unsigned int aper_i = 0; aper_i < std::min(name_size, atmos::indices_size); ++aper_i ) {
		fstm << "# Aper\tAlt.\t" << std::fixed << std::setprecision(2);
		for( vector_type::size_type j = 0; j < smooth_grid.size(); ++j ){
			fstm << smooth_grid(j) << "\t";
		}
		fstm << std::endl;
		for( vector_type::size_type i = 0; i < grid.size(); ++i ){
			boost::function1 <boost::tuple<vector_type,vector_type,vector_type>, double > alt_funct = function(grid(i));
			fstm << name[aper_i] << "\t";
			fstm << std::fixed << std::setprecision(3) << grid(i) << "\t";
			fstm << std::scientific << std::setprecision(4);
			for( vector_type::size_type j = 0; j < smooth_grid.size(); ++j ){
				fstm << alt_funct( smooth_grid(j) ).get<0>()(aper_i)/wscale << "\t";
			}
			fstm << std::endl;
		}
	}
}

factory::factory(const boost::filesystem::path& wd, const boost::filesystem::path& mass_response, const boost::filesystem::path& ccd_response ):
	m_mass_response(mass_response),
	m_ccd_response(ccd_response),
	m_wdir(wd),
	m_basis_grid( lg_generator(0.0, 0.04, 0.25), 50 ),
	m_smooth_grid( lg_generator(0.0, 0.1, 0.8), 8),
	m_nullexposure_factory( nullexposure_weight_creator(*this) ),
	m_shortexposure_factory( shortexposure_weight_creator(*this) ),
	m_longexposure_factory( longexposure_weight_creator(*this) ),
	m_variableexposure_factory( variableexposure_weight_creator(*this) ),
	m_mass_reaction( load_reaction(mass_response) ),
	m_dimm_reaction( load_reaction(ccd_response) ) {

	if( ! (boost::filesystem::exists( wd ) && boost::filesystem::is_directory( wd )) )
		throw std::runtime_error( wd.string() + " is not a directory or doesn't exists");
}
factory::~factory() {
}
const boost::filesystem::path factory::get_spectrum_filename( const std::string& spectype ) const {
	std::string spectral_type( spectype );
	boost::to_lower( spectral_type );
	boost::filesystem::path spect( m_wdir / (spectral_type + ".sp") );
	return spect;
}

const factory::nullexposure_weight_type& factory::nullexposure_weight( const std::string& spectype, const vector<double>& inner_diam, const vector<double>& outer_diam, double dimmbase, double dimmsize ) {
	return m_nullexposure_factory(spectype, inner_diam, outer_diam, dimmbase, dimmsize);
}
const factory::shortexposure_weight_type& factory::shortexposure_weight( const std::string& spectype, const vector<double>& inner_diam, const vector<double>& outer_diam ) {
	return m_shortexposure_factory(spectype, inner_diam, outer_diam);
}
const factory::longexposure_weight_type& factory::longexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam ) {
	return m_longexposure_factory(spectype, inner_diam, outer_diam);
}
const factory::variableexposure_weight_type& factory::variableexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam ) {
	return m_variableexposure_factory(spectype, inner_diam, outer_diam);
}

}}
