/*
    $Id: weight_factory.h 235 2011-07-17 16:10:52Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _SPECTRAL_FACTORY_H
#define _SPECTRAL_FACTORY_H

#include <map>
#include <string>

#include <boost/filesystem/path.hpp>
#include <boost/function.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "caching_factory.h"
#include "spline.h"
#include "local_context.h"
#include "nocase_less.h"
#include "vector_generator.h"
#include "utils.h"
#include "weif.h"

namespace atmos {
namespace weight {

using namespace boost::numeric::ublas;
using namespace utils::algo;
using namespace utils::ublas;
using utils::local_context;

std::pair< vector_type, vector_type > load_reaction( const boost::filesystem::path& file );
std::pair< vector_type, vector_type > photon_distribution( const vector_type& spectral_lambda, const vector_type& spectral_response, const vector_type& device_lambda, const vector_type& device_response );

void dump_weight( const boost::filesystem::path& file, const boost::function1< vector_type, double >& function, const vector_type& grid = vector_generator< lg_generator >( lg_generator(0.0, 0.04, 0.25), 49 ) );
void dump_variable_weight( const boost::filesystem::path& file, const boost::function1< boost::function1< boost::tuple<vector_type,vector_type,vector_type>, double >, double >& function, const vector_type& grid = vector_generator< lg_generator >( lg_generator(0.0, 0.04, 0.25), 49 ) );

typedef vector< vector_type > vector_vector_type;

class factory;

namespace detail {
	struct weight_wrapper: public local_context<const factory> {
		typedef boost::function1< vector_type, value_type > weight_type;
		typedef vector_type result_type;
		
		weight_wrapper( context_reference_type ctx, const weight_type& weight );
		
		result_type operator() (const value_type& x) const;
	private:
		weight_type weight_;
	};
	struct variableexposure_weight_wrapper: public local_context<const factory> {
		typedef boost::function2< vector_type, value_type, value_type > weight_type;
		typedef boost::function1< boost::tuple<vector_type,vector_type,vector_type>, value_type > result_type;
		
		variableexposure_weight_wrapper( context_reference_type ctx, const weight_type& weight );
		
		result_type operator() (const value_type& x) const;
	private:
		std::vector< boost::function1< vector_type, value_type > > weight_;
	};
	template<class T> struct mass_weight_creator: public local_context<const factory> {
		typedef weight_wrapper result_type;
		typedef T weight_type;
	
		mass_weight_creator( context_reference_type ctx ): local_context_type(ctx) {
		}
		
		result_type operator() ( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam ) const ;
	};
	template<class T> struct mass_variableexposure_weight_creator: public local_context<const factory> {
		typedef variableexposure_weight_wrapper result_type;
		typedef T weight_type;
	
		mass_variableexposure_weight_creator( context_reference_type ctx ): local_context_type(ctx) {
		}
		
		result_type operator() ( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam ) const ;
	};
	template<class T> struct massdimm_weight_creator: public mass_weight_creator<T> {
		typedef typename mass_weight_creator<T>::context_reference_type context_reference_type;
		typedef typename mass_weight_creator<T>::result_type result_type;
		typedef typename mass_weight_creator<T>::weight_type weight_type;
		
		massdimm_weight_creator( context_reference_type ctx ): mass_weight_creator<T>(ctx) {
		}
		
		result_type operator() ( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam, double dimmbase, double dimmsize ) const;
	};
	struct mass_weight_key {
		mass_weight_key( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam );
	
		bool operator==( const mass_weight_key& key ) const;
		bool operator!=( const mass_weight_key& key ) const;
		bool operator< ( const mass_weight_key& key ) const;
		
		std::ostream& print( std::ostream& stm ) const;
	private:
		const std::string spectype;
		const vector_type inner_diam;
		const vector_type outer_diam;
		
		nocase_less< std::string > nocase_less_;
	};
	struct massdimm_weight_key: public mass_weight_key {
		massdimm_weight_key( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam, double dimmbase, double dimmsize );
		
		bool operator==( const massdimm_weight_key& key ) const;
		bool operator!=( const massdimm_weight_key& key ) const;
		bool operator< ( const massdimm_weight_key& key ) const;
	private:
		double dimmbase;
		double dimmsize;
	};
	
	std::ostream& operator<< ( std::ostream& stm, const mass_weight_key& key );
	std::ostream& operator<< ( std::ostream& stm, const massdimm_weight_key& key );
}

class factory {
	friend class detail::weight_wrapper;
	friend class detail::variableexposure_weight_wrapper;
	friend class detail::massdimm_weight_creator< massdimm_weight >;
	friend class detail::mass_weight_creator< mass_shortexposure_weight >;
	friend class detail::mass_weight_creator< mass_longexposure_weight >;
	friend class detail::mass_variableexposure_weight_creator< mass_variableexposure_weight >;
public:
	typedef detail::massdimm_weight_creator< massdimm_weight >       nullexposure_weight_creator;
	typedef detail::mass_weight_creator< mass_shortexposure_weight > shortexposure_weight_creator;
	typedef detail::mass_weight_creator< mass_longexposure_weight >  longexposure_weight_creator;
	typedef detail::mass_variableexposure_weight_creator< mass_variableexposure_weight > variableexposure_weight_creator;
	
	typedef detail::weight_wrapper nullexposure_weight_type;
	typedef detail::weight_wrapper shortexposure_weight_type;
	typedef detail::weight_wrapper longexposure_weight_type;
	typedef detail::variableexposure_weight_wrapper variableexposure_weight_type;
	
	typedef detail::massdimm_weight_key nullexposure_key;
	typedef detail::mass_weight_key     shortexposure_key;
	typedef detail::mass_weight_key     longexposure_key;
	typedef detail::mass_weight_key     variableexposure_key;
	
	typedef caching_factory< nullexposure_weight_type, nullexposure_key, nullexposure_weight_creator >    nullexposure_factory_type;
	typedef caching_factory< shortexposure_weight_type, shortexposure_key, shortexposure_weight_creator > shortexposure_factory_type;
	typedef caching_factory< longexposure_weight_type, longexposure_key, longexposure_weight_creator >    longexposure_factory_type;
	typedef caching_factory< variableexposure_weight_type, variableexposure_key, variableexposure_weight_creator > variableexposure_factory_type;
private:
	boost::filesystem::path m_mass_response;
	boost::filesystem::path m_ccd_response;
	boost::filesystem::path m_wdir;

	vector_generator< lg_generator > m_basis_grid;
	vector_generator< lg_generator > m_smooth_grid;

	nullexposure_factory_type  m_nullexposure_factory;
	shortexposure_factory_type m_shortexposure_factory;
	longexposure_factory_type  m_longexposure_factory;
	variableexposure_factory_type m_variableexposure_factory;

	std::pair< vector_type, vector_type > m_mass_reaction;
	std::pair< vector_type, vector_type > m_dimm_reaction;
private:
	const boost::filesystem::path get_spectrum_filename( const std::string& spectype ) const;
public:
	factory(const boost::filesystem::path& wd, const boost::filesystem::path& mass_response, const boost::filesystem::path& ccd_response );
	~factory();

	const nullexposure_weight_type&  nullexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam, double dimmbase, double dimmsize );
	const shortexposure_weight_type& shortexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam );
	const longexposure_weight_type&  longexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam );
	const variableexposure_weight_type& variableexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam );

	const nullexposure_factory_type::const_iterator  nullexposure_begin() const { return m_nullexposure_factory.begin(); }
	const nullexposure_factory_type::const_iterator  nullexposure_end()   const { return m_nullexposure_factory.end(); }
	const shortexposure_factory_type::const_iterator shortexposure_begin() const { return m_shortexposure_factory.begin(); }
	const shortexposure_factory_type::const_iterator shortexposure_end()   const { return m_shortexposure_factory.end(); }
	const longexposure_factory_type::const_iterator  longexposure_begin() const { return m_longexposure_factory.begin(); }
	const longexposure_factory_type::const_iterator  longexposure_end()   const { return m_longexposure_factory.end(); }
	const variableexposure_factory_type::const_iterator variableexposure_begin() const { return m_variableexposure_factory.begin(); }
	const variableexposure_factory_type::const_iterator  variableexposure_end()   const { return m_variableexposure_factory.end(); }
};

namespace detail {
template<class T> typename mass_weight_creator<T>::result_type mass_weight_creator<T>::operator() ( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam ) const {
	typedef cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > spline_type;
	
	std::pair< vector_type, vector_type > spectral_response = load_reaction( context().get_spectrum_filename( spectype ) );
	std::pair< vector_type, vector_type > mass_response = photon_distribution( spectral_response.first, spectral_response.second, context().m_mass_reaction.first, context().m_mass_reaction.second );

	return result_type( context(), spline_type( context().m_basis_grid, utils::ublas::project( context().m_basis_grid, weight_type( mass_response.first, mass_response.second, inner_diam, outer_diam) ) ) );
};
template<class T> typename mass_variableexposure_weight_creator<T>::result_type mass_variableexposure_weight_creator<T>::operator() ( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam ) const {
	typedef cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > spline_type;
	
	std::pair< vector_type, vector_type > spectral_response = load_reaction( context().get_spectrum_filename( spectype ) );
	std::pair< vector_type, vector_type > mass_response = photon_distribution( spectral_response.first, spectral_response.second, context().m_mass_reaction.first, context().m_mass_reaction.second );

	return result_type( context(), weight_type( mass_response.first, mass_response.second, inner_diam, outer_diam ) );
};
template<class T> typename massdimm_weight_creator<T>::result_type massdimm_weight_creator<T>::operator() ( const std::string& spectype, const vector_type& inner_diam, const vector_type& outer_diam, double dimmbase, double dimmsize ) const {
	typedef cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > spline_type;
	
	std::pair< vector_type, vector_type > spectral_response = load_reaction( this->context().get_spectrum_filename( spectype ) );
	std::pair< vector_type, vector_type > mass_response = photon_distribution( spectral_response.first, spectral_response.second, this->context().m_mass_reaction.first, this->context().m_mass_reaction.second );
	std::pair< vector_type, vector_type > dimm_response = photon_distribution( spectral_response.first, spectral_response.second, this->context().m_dimm_reaction.first, this->context().m_dimm_reaction.second );

	return result_type( this->context(), spline_type( this->context().m_basis_grid, utils::ublas::project( this->context().m_basis_grid, weight_type( mass_response.first, mass_response.second, dimm_response.first, dimm_response.second, inner_diam, outer_diam, dimmsize, dimmbase ) ) ) );
};
}


}}

#endif //_SPECTRAL_FACTORY_H
