#define BOOST_TEST_MODULE variable_exposure_test
#include <boost/test/included/unit_test.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>

#include "../../src/atmos.h"
#include "../../src/weight_factory.h"
#include "../../src/weif.h"
#include "../../src/variableexposure_profile.h"

#include <string>

using namespace boost::numeric::ublas;

typedef vector<double> vector_type;

namespace {

const std::string spectral_dir = "spectra";
const std::string mass_response = "mass_green_rcx.crv";
const std::string ccd_response = "ccd.crv";
const std::string spectral_class = "a05";

std::pair<vector_type, vector_type> init_mass_geometry() {
	vector_type inner_diam(atmos::channels);
	vector_type outer_diam(atmos::channels);
	
	inner_diam(0)=0.0;outer_diam(0)=1.27;
	inner_diam(1)=1.3;outer_diam(1)=2.15;
	inner_diam(2)=2.2;outer_diam(2)=3.85;
	inner_diam(3)=3.9;outer_diam(3)=5.50;
	inner_diam = inner_diam * 14.5*0.1;
	outer_diam = outer_diam * 14.5*0.1;
	
	return std::make_pair( inner_diam, outer_diam );
}

vector_type init_grid() {
	vector_type grid(6);
	grid(0) = 0.5;
	grid(1) = 1;
	grid(2) = 2;
	grid(3) = 4;
	grid(4) = 8;
	grid(5) = 16;
	
	return grid;
}

vector_type init_extended_grid() {
	return utils::ublas::vector_generator< atmos::lg_generator >( atmos::lg_generator(0.0, 0.04, 0.25), 49 );
}

}

BOOST_AUTO_TEST_CASE(zero_wind) {
	std::pair<vector_type, vector_type> diam = init_mass_geometry();
	
	atmos::weight::factory weight_factory( spectral_dir, mass_response, ccd_response );
	
	typedef atmos::weight::factory::variableexposure_weight_type variableexposure_weight_type;
	variableexposure_weight_type varexp_weight = weight_factory.variableexposure_weight( spectral_class, diam.first, diam.second );
	
	const vector_type grid = init_grid();
	const unsigned int exp_num = 7;
	symmetric_matrix<double> covariance = identity_matrix<double>(exp_num * atmos::indices_size);
	
	atmos::variableexposure::profile::functor profile_functor( grid, covariance, 1.0, varexp_weight );
	
	const vector_type extended_grid = init_extended_grid();
	const vector_type cn2_init = scalar_vector<double>( grid.size(), 1e-14 );
	const vector_type windspeed_init = scalar_vector<double>( grid.size(), 1.0 );
	
	for( vector_type::const_iterator it = extended_grid.begin(); it != extended_grid.end(); ++it ) {
		variableexposure_weight_type::result_type local_weight = varexp_weight(*it);
		
		vector_type indices(exp_num * atmos::indices_size);
		double tau = 1e-3;
		for( unsigned int i = 0; i<exp_num; ++i, tau*=2 ) {
			vector_type tmp;
			boost::tie(tmp, boost::tuples::ignore, boost::tuples::ignore) = local_weight(tau * 0.0);
			subrange(indices, atmos::indices_size*i, atmos::indices_size*(i+1) ) = tmp * 1e-13;
		}
		
		const atmos::variableexposure::profile::functor::result_type& ret = profile_functor(indices,cn2_init,windspeed_init);
		std::cout << *it << " " << ( ret.get<0>() ? "success " :"fail " ) << "\t" << ret.get<1>() << "\t" << ret.get<3>() << "\t" << ret.get<5>() << std::endl;
	}
}
