#define BOOST_TEST_MODULE weif_test
#include <boost/test/included/unit_test.hpp>
#include <boost/numeric/ublas/io.hpp>

#include "../../src/atmos.h"
#include "../../src/weight_factory.h"
#include "../../src/weif.h"
#include "../../src/utils.h"

BOOST_AUTO_TEST_CASE( wind_factor ) {
	using atmos::weight::wind_factor;

	BOOST_CHECK_CLOSE( wind_factor(0.0),  1.0          , 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(0.25), 0.9033059998 , 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(0.5),  0.6766701410 , 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(0.75), 0.4503268909 , 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(1.0),  0.3092552256 , 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(2.0),  0.1573429508 , 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(4.0),  0.07923837253, 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(5.0),  0.06346579361, 2e-0);
	BOOST_CHECK_CLOSE( wind_factor(10.0), 0.03179559908, 2e-0);
}

double smooth_limit( double height ) {
	return 2 * std::sqrt( height * 1000 * 5e-6 ) * 100;
}

BOOST_AUTO_TEST_CASE( weight_smoothed_test ) {
	using namespace utils::ublas;
	using namespace atmos;
		
	typedef weight::vector_type vector_type;
	typedef vector< vector_type > vector_vector_type;
	typedef utils::ublas::cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > spline_type;
	std::pair< vector_type, vector_type > mass_reaction = weight::load_reaction("./mass_maid.crv");
	std::pair< vector_type, vector_type > spectral_response = weight::load_reaction("./a05.sp");
	std::pair< vector_type, vector_type > mass_response = weight::photon_distribution( spectral_response.first, spectral_response.second, mass_reaction.first, mass_reaction.second );
	
	vector_type inner_diam(atmos::channels);
	vector_type outer_diam(atmos::channels);
	inner_diam(0)=0.0;outer_diam(0)=1.27;
	inner_diam(1)=1.3;outer_diam(1)=2.15;
	inner_diam(2)=2.2;outer_diam(2)=3.85;
	inner_diam(3)=3.9;outer_diam(3)=5.50;
	inner_diam = inner_diam * 14.5*0.1;
	outer_diam = outer_diam * 14.5*0.1;

	atmos::weight::mass_variableexposure_weight wf( mass_response.first, mass_response.second, inner_diam, outer_diam );
	
	vector_type height_grid = vector_generator< atmos::lg_generator >( atmos::lg_generator(0.0, 0.04, 0.25), 50 );
	vector_type smooth_grid = vector_generator< atmos::lg_generator >( atmos::lg_generator(0.0, 0.1, 0.8), 8);

	std::cerr << "# height: " << height_grid << std::endl;
	std::cerr << "# smooth: " << smooth_grid << std::endl;
	
	for( size_type i = 0; i < height_grid.size(); ++i ) {
		std::cerr << height_grid(i) << "\t";
		double smooth_lim = smooth_limit(height_grid(i));
		std::cerr << "# smooth: " << smooth_lim * smooth_grid << std::endl;
		
		for( size_type j = 0; j < smooth_grid.size(); ++j ) {
			std::cerr << wf( height_grid(i), smooth_grid(j)*smooth_lim )(0) << "\t";
		}
		std::cerr << std::endl;
	}
}

BOOST_AUTO_TEST_CASE( weight_longexposure_test ) {
	using namespace utils::ublas;
	using namespace atmos;
	
	typedef weight::vector_type vector_type;
	std::pair< vector_type, vector_type > mass_reaction = weight::load_reaction("./mass_maid.crv");
	std::pair< vector_type, vector_type > spectral_response = weight::load_reaction("./a05.sp");
	std::pair< vector_type, vector_type > mass_response = weight::photon_distribution( spectral_response.first, spectral_response.second, mass_reaction.first, mass_reaction.second );
	
	vector_type inner_diam(atmos::channels);
	vector_type outer_diam(atmos::channels);
	inner_diam(0)=0.0;outer_diam(0)=1.27;
	inner_diam(1)=1.3;outer_diam(1)=2.15;
	inner_diam(2)=2.2;outer_diam(2)=3.85;
	inner_diam(3)=3.9;outer_diam(3)=5.50;
	inner_diam = inner_diam * 14.5*0.1;
	outer_diam = outer_diam * 14.5*0.1;

	atmos::weight::mass_longexposure_weight wf( mass_response.first, mass_response.second, inner_diam, outer_diam );
	
	vector_type height_grid = vector_generator< atmos::lg_generator >( atmos::lg_generator(0.0, 0.04, 0.25), 50 );
	
	for( vector_type::size_type h = 0; h < height_grid.size(); ++h ) {
		std::cout << height_grid(h) << "\t" << wf(height_grid(h))(0) << "\n";
	}
	std::cout << std::endl;
}

BOOST_AUTO_TEST_CASE( weight_shortexposure_test ) {
	using namespace utils::ublas;
	using namespace atmos;
	
	typedef weight::vector_type vector_type;
	std::pair< vector_type, vector_type > mass_reaction = weight::load_reaction("./mass_maid.crv");
	std::pair< vector_type, vector_type > spectral_response = weight::load_reaction("./a05.sp");
	std::pair< vector_type, vector_type > mass_response = weight::photon_distribution( spectral_response.first, spectral_response.second, mass_reaction.first, mass_reaction.second );
	
	vector_type inner_diam(atmos::channels);
	vector_type outer_diam(atmos::channels);
	inner_diam(0)=0.0;outer_diam(0)=1.27;
	inner_diam(1)=1.3;outer_diam(1)=2.15;
	inner_diam(2)=2.2;outer_diam(2)=3.85;
	inner_diam(3)=3.9;outer_diam(3)=5.50;
	inner_diam = inner_diam * 14.5*0.1;
	outer_diam = outer_diam * 14.5*0.1;

	atmos::weight::mass_shortexposure_weight wf( mass_response.first, mass_response.second, inner_diam, outer_diam );
	
	vector_type height_grid = vector_generator< atmos::lg_generator >( atmos::lg_generator(0.0, 0.04, 0.25), 50 );
	
	for( vector_type::size_type h = 0; h < height_grid.size(); ++h ) {
		std::cout << wf(height_grid(h))(0) << "\t";
	}
	std::cout << std::endl;
}
