#ifndef _TRM_CGM_H
#define _TRM_CGM_H

#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/vector.hpp>

namespace trm {
namespace cgm {

using namespace boost::numeric::ublas;

namespace precond {
namespace detail {
template<class M, class T> symmetric_matrix< typename M::value_type > ssor( const M& m, const T& omega ) {
	typedef typename M::value_type value_type;
	typedef typename M::size_type size_type;
	
	using std::sqrt;
	using std::abs;
	
	const size_type size = BOOST_UBLAS_SAME( m.size1(), m.size2() );
	
	if( size < 1 )
		return symmetric_matrix< value_type >(0,0);

	banded_matrix< value_type > diag(size,size);
	
	value_type max_diag(0);
	for( size_type i = 0; i < size; ++i ) {
		if( abs(m(i,i)) > max_diag ) {
			max_diag = abs(m(i,i));
		}
	}
	
	for( size_type i = 0; i < size; ++i ) {
		if( abs(m(i,i)) > 10 * std::numeric_limits<value_type>::epsilon() * max_diag ) {
			diag(i,i) = 1.0 / sqrt(abs(m(i,i)));
		} else {
			diag(i,i) = 0;
		}
	}
	
	matrix< value_type > rmtx = diag;
	triangular_adaptor< const M, strict_lower > lower(m);
	
	inplace_solve( omega*prod( diag, triangular_matrix< value_type, strict_lower >(prod(lower, diag)) ), rmtx, unit_lower_tag() );

	return prod( trans(rmtx), rmtx );
}
}

template<class T> symmetric_matrix<T> ssor( const symmetric_matrix<T>& m, const T& omega ) {
	return detail::ssor(m, omega);
}
template<class M,class T> symmetric_matrix< typename M::value_type > ssor( const symmetric_adaptor<M>& m, const T& omega ) {
	return detail::ssor(m, omega);
}
template<class T> symmetric_matrix<T> ssor( const symmetric_matrix<T>& m ) {
	return detail::ssor(m, 1.0);
}
template<class M> symmetric_matrix< typename M::value_type > ssor( const symmetric_adaptor<M>& m ) {
	return detail::ssor(m, 1.0);
}

}

namespace detail {
template<class T, class P>
boost::tuple< vector<T>, T, T > truncated_cgm( const vector<T>& grad, const symmetric_matrix<T>& hesse, const T& region, const P& precond ) {
	// FIXME:
	// we probably need default value of region = inf for ordinary CGM
	// we need function : vector<double> -> vector<double> instead of hesse ( prod(hesse,_1) )
	// we need function : vector<double> -> vector<double> for preconditioner ( prod(preconditioner,_1) )
	typedef T value_type;
	typedef vector<T> vector_type;
	typedef typename vector_type::size_type size_type;

	assert( grad.size() == hesse.size1() );
	assert( grad.size() == hesse.size2() );
	
	vector_type s0 = zero_vector<value_type>(grad.size());
	vector_type g0 = grad;
	vector_type v0 = /*preconditioner*/prod(precond,g0);
	vector_type p0 = -v0;
	
	value_type alpha = 0;
	
	const value_type delta2 = region*region;
	value_type inner_g0v0 = inner_prod(g0,v0);
	value_type norm2_s0 = 0;
	value_type inner_s0_Mp0 = 0;
	value_type norm2_p0 = inner_g0v0;
	
	const value_type norm2_g0_limit = norm_2(grad)*std::min(0.1,std::sqrt(norm_2(grad)));
	size_type iter_limit = grad.size();
	value_type decrease = 0;
	
	do {
		value_type curvature = inner_prod(p0,prod(hesse,p0));
		value_type alpha = inner_g0v0/curvature;
		if( curvature > 0 && (norm2_s0 + 2*alpha*inner_s0_Mp0 + alpha*alpha*norm2_p0) < delta2 ) {
			s0 = s0 + alpha*p0;
			g0 = g0 + alpha*prod(hesse,p0);
			v0 = /*preconditioner*/prod(precond,g0);
				
			value_type inner_gv = inner_prod(g0,v0);
			value_type beta = inner_gv/inner_g0v0;
			p0 = -v0 + beta*p0;
				
			norm2_s0 = norm2_s0 + 2*alpha*inner_s0_Mp0 + alpha*alpha*norm2_p0;
			inner_s0_Mp0 = beta*(inner_s0_Mp0 + alpha*norm2_p0);
			norm2_p0 = inner_gv + beta*beta*norm2_p0;
			decrease -= 0.5*alpha*inner_g0v0;
			inner_g0v0 = inner_gv;
		} else {
			if( norm2_p0 > 0 ) {
				value_type sigma = (-inner_s0_Mp0+std::sqrt(inner_s0_Mp0*inner_s0_Mp0+norm2_p0*(delta2-norm2_s0)))/norm2_p0;
				s0 = s0 + sigma*p0;
				norm2_s0 = norm2_s0 + 2*sigma*inner_s0_Mp0 + sigma*sigma*norm2_p0;
				decrease -= sigma*inner_g0v0 - 0.5*sigma*sigma*curvature;
			}
			break;
		}
		--iter_limit;
	} while( norm_2(g0) > norm2_g0_limit && iter_limit > 0 );
	
	return boost::make_tuple(s0, std::sqrt(norm2_s0), decrease);
}
}

template<class T> boost::tuple< vector<T>, T, T > truncated_cgm( const vector<T>& grad, const symmetric_matrix<T>& hesse, const T& region, const matrix<T>& precond ) {
	return detail::truncated_cgm( grad, hesse, region, precond );
}
template<class T> boost::tuple< vector<T>, T, T > truncated_cgm( const vector<T>& grad, const symmetric_matrix<T>& hesse, const T& region, const symmetric_matrix<T>& precond ) {
	return detail::truncated_cgm( grad, hesse, region, precond );
}
template<class T> boost::tuple< vector<T>, T, T > truncated_cgm( const vector<T>& grad, const symmetric_matrix<T>& hesse, const T& region ) {
	return detail::truncated_cgm( grad, hesse, region, matrix<T>(identity_matrix<T>(hesse.size1(),hesse.size2())) );
}

}}

#endif // _TRM_CGM_H
