#ifndef _TRM_DELTA_H
#define _TRM_DELTA_H

#include <cmath>

namespace trm {
namespace delta {

/** @brief function for rho-ratio calculation
 *  @param[in] trial_f Value of \f$ f_{trial} \f$
 *  @param[in] initial_f Value of initial \f$ f_{0} \f$
 *  @param[in] diff_m Value of model difference \f$ \delta m \f$
 *  @return \f$ \rho \f$
 * 
 *  The common used quantity is the
 *  \f$ \rho \equiv \frac{f_{trial} - f_{0}}{\delta m} \f$. Model difference
 *  \f$ \delta m \f$ could be obtained as ready value from specifil algorithm or
 *  as difference between value of new model and initial model.
 * 
 */
template<class T> T rho_ratio( const T& trial_f, const T& initial_f, const T& diff_m ) {
	/* This function requires futher consideration. 
	 * */
	typedef T value_type;
	
	using std::max;
	using std::abs;

	value_type diff_f = trial_f - initial_f;
	
	if( abs(diff_f) < 10 * std::numeric_limits<value_type>::epsilon() * trial_f ) {
		return 1.0;
	}
	if( max(abs(diff_f),abs(diff_m)) < 10 * std::numeric_limits<value_type>::epsilon() ) {
		return 1.0;
	}
	return diff_f / diff_m;
}
	
namespace initial {
	template<class Context> struct tenth_grad {
		typedef typename Context::value_type value_type;
		typedef typename Context::grad_type  grad_type;
		typedef typename Context::hesse_type hesse_type;
			
		value_type initial_delta( Context& context ) const {
			grad_type grad;
			
			boost::tuples::tie(boost::tuples::ignore, grad, boost::tuples::ignore) = context.evaluate_function(context.initial_x());
			
			return 0.1 * norm_2(grad);
		}
	};
	template<class Context> struct nearest_cauchy {
		typedef typename Context::value_type value_type;
		typedef typename Context::grad_type  grad_type;
		typedef typename Context::hesse_type hesse_type;

		value_type initial_delta( Context& context ) const {
			using std::abs;
			using std::pow;
			
			grad_type grad;
			hesse_type hesse;
			
			boost::tuples::tie(boost::tuples::ignore, grad, hesse) = context.evaluate_function( context.initial_x() );
			
			const value_type epsilonM = 10 * std::numeric_limits<value_type>::epsilon();
			value_type curvature = inner_prod(grad, prod(hesse, grad));
			value_type slope = inner_prod(grad, grad);
			
			return abs((pow(slope,1.5)+epsilonM) / (curvature+epsilonM));
		}
	};
}
namespace update {
	template<class Context> struct simple_update {
		typedef typename Context::value_type value_type;
		
		/* defaults are from [TRM] book */
		simple_update( value_type n1 = 0.05, value_type n2 = 0.9, value_type a1 = 2.5, value_type a2 = 0.25 ): n1_(n1), n2_(n2), a1_(a1), a2_(a2) {
		}
		
		value_type update_delta( const Context& context ) const {
			using std::max;
			
			value_type step_norm = context.last_step_norm();
			value_type ratio = context.last_rho_ratio();
			value_type delta = context.last_delta();
			
			if(ratio < n1_) delta = a2_*step_norm;
			else if (ratio < n2_) delta = delta;
			else delta = max( delta, a1_*step_norm );
			
			return delta;
		}
		
		const value_type& rho_ratio_threshold() const { return n1_; }
	private:
		value_type n1_,n2_,a1_,a2_;
	};
}
}
}

#endif // _TRM_DELTA_H
