#ifndef _TRM_FUNCTOR_H
#define _TRM_FUNCTOR_H

#include <boost/function.hpp>
#include <boost/tuple/tuple.hpp>

#include "traits.h"

namespace trm {
namespace functor {
	template<class T> struct basic {
		typedef T value_type;
		typedef typename type_traits<T>::vector_type vector_type;
		typedef typename type_traits<T>::symmetric_matrix_type symmetric_matrix_type;
	
		typedef typename boost::tuple< value_type, vector_type, symmetric_matrix_type > result_type;
		typedef vector_type arity_type;
		typedef boost::function1< result_type, arity_type > function_type;
		
		basic( const function_type& nested ): nested_(nested) {}
		
		result_type operator() ( const arity_type& x ) {
			return nested_(x);
		}
	private:
		function_type nested_;
	};
	template<class T> struct least_squares {
		typedef T value_type;
		typedef typename type_traits<T>::vector_type vector_type;
		typedef typename type_traits<T>::matrix_type matrix_type;
		typedef typename type_traits<T>::symmetric_matrix_type symmetric_matrix_type;
	
		typedef typename boost::tuple< value_type, vector_type, symmetric_matrix_type > result_type;
		typedef vector_type arity_type;
		typedef boost::function1< boost::tuple<vector_type, matrix_type, symmetric_matrix_type>, arity_type > function_type;
		
		least_squares( const function_type& nested ): nested_(nested) {}
		
		result_type operator() ( const arity_type& x ) {
			using namespace boost::numeric::ublas;

			vector_type residual;
			matrix_type jacobian;
			symmetric_matrix_type hesse;
			boost::tie(residual,jacobian,hesse) = nested_(x);
			return boost::make_tuple( 0.5*inner_prod(residual,residual), prod(trans(jacobian),residual), prod(trans(jacobian),jacobian)+hesse );
		}
	private:
		function_type nested_;
	};
	template<class T> struct regularization {
		typedef T value_type;
		typedef typename type_traits<T>::vector_type vector_type;
		typedef typename type_traits<T>::symmetric_matrix_type symmetric_matrix_type;
	
		typedef typename boost::tuple< value_type, vector_type, symmetric_matrix_type > result_type;
		typedef vector_type arity_type;
		typedef boost::function1< result_type, arity_type > function_type;
	
		regularization( const function_type& nested, const function_type& smooth ): nested_(nested), smooth_(smooth) {}
		
		result_type operator() ( const arity_type& x ) {
			using namespace boost::tuples;
			
			const typename function_type::result_type& nested = nested_(x);
			const typename function_type::result_type& smooth = smooth_(x);
			
			return boost::make_tuple( get<0>(nested)+get<0>(smooth), get<1>(nested)+get<1>(smooth), get<2>(nested)+get<2>(smooth) );
		}
	private:
		function_type nested_,smooth_;
	};
	template<class T> struct vector_norm {
		typedef T value_type;
		typedef typename type_traits<T>::vector_type vector_type;
		typedef typename type_traits<T>::symmetric_matrix_type symmetric_matrix_type;
		
		typedef typename boost::tuple< value_type, vector_type, symmetric_matrix_type > result_type;
		typedef vector_type arity_type;
		
		vector_norm( const value_type& scale, const typename boost::numeric::ublas::zero_vector<value_type>::size_type& size ): scale_(scale), psi_( typename boost::numeric::ublas::zero_vector<value_type>(size) ) {}
		vector_norm( const value_type& scale, const vector_type& psi ): scale_(scale), psi_(psi) {}
		
		result_type operator() ( const arity_type& x ) {
			using namespace boost::numeric::ublas;
			
			assert( x.size() == psi_.size() );
			
			return boost::make_tuple( scale_*inner_prod(x-psi_,x-psi_), (x-psi_)*2*scale_, identity_matrix< value_type >( x.size() )*2*scale_ );
		}
		
	private:
		value_type scale_;
		vector_type psi_;
	};
}
}

#endif // _TRM_FUNCTOR_H
