#ifndef _TRM_STEIHAUG_TOINT_H
#define _TRM_STEIHAUG_TOINT_H

#include <boost/numeric/ublas/storage.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/triangular.hpp>

#include <boost/numeric/ublas/io.hpp>

#include <boost/tuple/tuple.hpp>

#include "cgm.h"
#include "delta.h"
#include "driver.h"
#include "functor.h"
#include "stop.h"
#include "traits.h"

namespace trm {
namespace detail {

using namespace boost::numeric::ublas;
	
class active_set {
public:
	typedef unbounded_array< std::size_t > array_type;
	typedef array_type::const_iterator const_iterator;
private:
	void swap_indexes( array_type& src, array_type& dest, std::size_t index ) {
		array_type::iterator it = std::find( src.begin(), src.end(), index );
		
		if( it != src.end() ) {
			dest.resize( dest.size()+1, index ); // push_back
			std::swap( *it, src[src.size()-1] );
			src.resize( src.size()-1, std::size_t() ); // pop_back
		}
	}
public:
	active_set( std::size_t size ): free_set_(size), zero_set_(0) {
		for( std::size_t i = 0; i < free_set_.size(); ++i ) {
			free_set_[i]=i;
		}
	}
	
	void lock_index( std::size_t index ) {
		swap_indexes( free_set_, zero_set_, index );
	}
	void free_index( std::size_t index ) {
		swap_indexes( zero_set_, free_set_, index );
	}
	template<class V> vector_indirect<V> make_free_vector_indirect( V& vec ) const {
		return vector_indirect<V>( vec, indirect_array<>(free_set_.size(),free_set_) );
	}
	template<class M> matrix_indirect<M> make_free_matrix_indirect( M& mtx ) const {
		return matrix_indirect<M>( mtx, indirect_array<>(free_set_.size(),free_set_), indirect_array<>(free_set_.size(),free_set_) );
	}
	template<class V> vector_indirect<V> make_zero_vector_indirect( V& vec ) const {
		return vector_indirect<V>( vec, indirect_array<>(zero_set_.size(),zero_set_) );
	}
	template<class M> matrix_indirect<M> make_zero_matrix_indirect( M& mtx ) const {
		return matrix_indirect<M>( mtx, indirect_array<>(zero_set_.size(),zero_set_), indirect_array<>(zero_set_.size(),zero_set_) );
	}
	
	const_iterator free_begin() const { return free_set_.begin(); }
	const_iterator free_end()   const { return free_set_.end(); }
	const_iterator zero_begin() const { return zero_set_.begin(); }
	const_iterator zero_end()   const { return zero_set_.end(); }
private:
	array_type free_set_,zero_set_;
};

}}

namespace trm {
namespace core {
template<class Context> struct steihaug_toint {
	typedef typename Context::value_type value_type;
	typedef typename Context::vector_type vector_type;
	typedef typename Context::function_type function_type;
	typedef typename Context::grad_type grad_type;
	typedef typename Context::hesse_type hesse_type;
		
	steihaug_toint( Context& context ): delta_(0), rho_ratio_(0), step_norm_(0), context_(context) {
		x_ = context.initial_x();
		trial_x_ = x_;
		boost::tie(function_, grad_, hesse_) = context.evaluate_function(x_);
	}
	void update_delta( value_type new_delta ) {
		delta_ = new_delta;
	}
	virtual void update_x( const vector_type& new_x ) {
		x_ = new_x;
		boost::tie(function_,grad_,hesse_) = context_.evaluate_function(x_);
	}
	virtual void iterate() {
		value_type model_decrease, trial_function;
		vector_type step;
		
		boost::tie(step,step_norm_,model_decrease) = trm::cgm::truncated_cgm(grad_, hesse_, delta_, trm::cgm::precond::ssor(this->hesse_) );
		trial_x_ = x_ + step;
		boost::tie(trial_function,boost::tuples::ignore,boost::tuples::ignore) = context_.evaluate_function( trial_x_ );
		
		rho_ratio_ = trm::delta::rho_ratio(trial_function,function_,model_decrease);
	}
	virtual bool check_stop() const {
		using namespace trm::stop;
		
		bool ret = false;
		
		ret = ret | context_.stop_strategy().check_step(step(),x_);
		ret = ret | context_.stop_strategy().check_delta(delta_,x_);
		ret = ret | context_.stop_strategy().check_grad(grad_,x_,function_);
		
		return ret;
	}
	
	const vector_type& x() const { return x_; }
	
	const value_type& function() const { return function_; }
	const grad_type&  grad()     const { return grad_; }
	const hesse_type& hesse()    const { return hesse_; }
	
	vector_type step() const { return trial_x_ - x_; }
	const vector_type& trial_x() const { return trial_x_; }
	value_type step_norm()    const { return step_norm_; }
	value_type rho_ratio()    const { return rho_ratio_; }
	value_type delta()        const { return delta_; }
private:

protected:
	value_type delta_;
	value_type step_norm_;
	value_type rho_ratio_;
	vector_type x_,trial_x_;
	value_type function_;
	grad_type grad_;
	hesse_type hesse_;
	
	Context& context_;
};

template<class Context> struct steihaug_toint_nonnegative: public steihaug_toint<Context> {
	typedef typename steihaug_toint<Context>::value_type value_type;
	typedef typename steihaug_toint<Context>::vector_type vector_type;
	typedef typename steihaug_toint<Context>::function_type function_type;
	typedef typename steihaug_toint<Context>::grad_type grad_type;
	typedef typename steihaug_toint<Context>::hesse_type hesse_type;
private:
	bool do_free_bounds( const vector_type& grad ) {
		value_type free_grad_norm2 = 0;
		for( detail::active_set::const_iterator it = active_set_.free_begin(); it != active_set_.free_end(); ++it ) {
			free_grad_norm2 += grad(*it)*grad(*it);
		}
		bool ret = false;
		std::size_t max_i = 0;
		value_type max_zero_grad = 0;
		for( detail::active_set::const_iterator it = active_set_.zero_begin(); it != active_set_.zero_end(); ++it ) {
			if( grad(*it) < 0 && std::abs(grad(*it)) > max_zero_grad ) {
				max_i = *it;
				max_zero_grad = std::abs(grad(*it));
				ret = true;
			}
		}
		if( ret && max_zero_grad*max_zero_grad > free_grad_norm2 ) {
			last_unlock_ = max_i;
			active_set_.free_index(max_i);
		} else {
			last_unlock_ = std::size_t(-1);
			ret = false;
		}
		return ret;
	}
	bool do_lock_bounds( const vector_type& new_x ) {
		for( detail::active_set::const_iterator it = active_set_.free_begin(); it != active_set_.free_end(); ++it ) {
			if( new_x(*it) <= 0 ) {
				active_set_.lock_index(*it);
				return true; // don't play with broken iterators
			}
		}
		return false;
	}
	value_type need_project( const vector_type& x, const vector_type& step ) const {
		using namespace boost::numeric::ublas;
		
		value_type alpha = 1.0;
		
		for( typename vector_type::size_type i = 0; i < BOOST_UBLAS_SAME(x.size(),step.size()); ++i ) {
			value_type tmp_alpha = -x(i)/step(i);
			if( tmp_alpha > 0.0 && alpha > tmp_alpha ) {
				alpha = tmp_alpha;
			}
		}
		return alpha;
	}
public:
	steihaug_toint_nonnegative(Context& context): steihaug_toint<Context>(context), active_set_(context.initial_x().size()), last_unlock_(-1) {
	}
	
	virtual void update_x( const vector_type& new_x ) {
		this->x_ = new_x;
		while( do_lock_bounds(this->x_) );
		boost::tie(this->function_,this->grad_,this->hesse_) = this->context_.evaluate_function(this->x_);
		do_free_bounds(this->grad_);
		this->grad_ = active_set_.make_free_vector_indirect( this->grad_ );
		this->hesse_ = active_set_.make_free_matrix_indirect( this->hesse_ );
	}
	virtual void iterate() {
		using namespace boost::numeric::ublas;
		
		value_type model_decrease, trial_function;
		
		vector_type new_step;
		boost::tie(new_step,this->step_norm_,model_decrease) = trm::cgm::truncated_cgm(this->grad_, this->hesse_, this->delta_, trm::cgm::precond::ssor(this->hesse_) );

		this->trial_x_ = this->x_;

		value_type alpha = need_project( active_set_.make_free_vector_indirect(this->x_), new_step );
		if( alpha > 0.0 && alpha < 1.0 ) {
			new_step = alpha * new_step;
			this->step_norm_ = alpha * this->step_norm_;
			model_decrease = inner_prod(this->grad_,new_step) + 0.5 * inner_prod(new_step,prod(this->hesse_,new_step));
		}
		active_set_.make_free_vector_indirect(this->trial_x_) = active_set_.make_free_vector_indirect(this->x_) + new_step;
		if( last_unlock_ != std::size_t(-1) && this->trial_x_(last_unlock_) < 0 ) {
			this->trial_x_(last_unlock_) = 0;
		}
		// rounding error
		for( typename vector_type::size_type i = 0; i < this->trial_x_.size(); ++i ) {
			if( this->trial_x_(i) < 0.0 ) this->trial_x_(i) = 0.0;
		}

		boost::tie(trial_function,boost::tuples::ignore,boost::tuples::ignore) = this->context_.evaluate_function( this->trial_x_ );
		
		this->rho_ratio_ = trm::delta::rho_ratio(trial_function,this->function_,model_decrease);
	}
	virtual bool check_stop() const {
		using namespace trm::stop;
		
		bool ret = false;
		
		ret = ret | this->context_.stop_strategy().check_step(this->step(),this->x_);
		ret = ret | this->context_.stop_strategy().check_delta(this->delta_,active_set_.make_free_vector_indirect(this->x_));
		ret = ret | this->context_.stop_strategy().check_grad(this->grad_,active_set_.make_free_vector_indirect(this->x_),this->function_);
		
		return ret;
	}
private:
	detail::active_set active_set_;
	std::size_t last_unlock_;
};
}
}

namespace trm {
	template<class T,class F> 
	boost::tuple< trm::stop::basic<T>, typename trm::type_traits<T>::function_type, typename trm::type_traits<T>::vector_type >
	steihaug_toint( const ublas::vector<T>& x0, F f, const trm::stop::basic<T>& stop_strategy ) {
		trm::basic_driver< T, trm::delta::initial::nearest_cauchy,
			trm::core::steihaug_toint,
			trm::delta::update::simple_update,
			trm::stop::basic<T> > dr(x0,f,stop_strategy);
		return dr.minimize();
	}
	template<class T,class F> 
	boost::tuple< trm::stop::basic<T>, typename trm::type_traits<T>::function_type, typename trm::type_traits<T>::vector_type >
	steihaug_toint_nonnegative( const ublas::vector<T>& x0, F f, const trm::stop::basic<T>& stop_strategy ) {
		trm::basic_driver< T, trm::delta::initial::nearest_cauchy,
			trm::core::steihaug_toint_nonnegative,
			trm::delta::update::simple_update,
			trm::stop::basic<T> > dr(x0,f,stop_strategy);
		return dr.minimize();
	}
	template<class T,class F> 
	boost::tuple< trm::stop::basic<T>, typename trm::type_traits<T>::function_type, typename trm::type_traits<T>::vector_type >
	steihaug_toint_least_squares( const ublas::vector<T>& x0, F f, const trm::stop::basic<T>& stop_strategy ) {
		trm::basic_driver< T, trm::delta::initial::nearest_cauchy,
			trm::core::steihaug_toint,
			trm::delta::update::simple_update,
			trm::stop::basic<T> > dr(x0,trm::functor::least_squares<T>(f),stop_strategy);
		return dr.minimize();
	}
	template<class T,class F> 
	boost::tuple< trm::stop::basic<T>, typename trm::type_traits<T>::function_type, typename trm::type_traits<T>::vector_type >
	steihaug_toint_nonnegative_least_squares( const ublas::vector<T>& x0, F f, const trm::stop::basic<T>& stop_strategy ) {
		trm::basic_driver< T, trm::delta::initial::nearest_cauchy,
			trm::core::steihaug_toint_nonnegative,
			trm::delta::update::simple_update,
			trm::stop::basic<T> > dr(x0,trm::functor::least_squares<T>(f),stop_strategy);
		return dr.minimize();
	}
}

#endif // _TRM_STEIHAUG_TOINT_H
