#ifndef _TEST_GAUSS_H
#define _TEST_GAUSS_H

#include <boost/tuple/tuple.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/vector.hpp>

namespace ublas = boost::numeric::ublas;

struct gauss {
	gauss( const ublas::vector<double>& center ): center_(center) {
	}
	
	boost::tuple< double, ublas::vector<double>, ublas::symmetric_matrix<double> > operator() (const ublas::vector<double>& x) const {
		ublas::vector<double> vec = x - center_;
		ublas::vector<double> exp(vec);
		for( unsigned int i = 0; i < exp.size(); ++i ) {
			exp(i) = std::exp( -vec(i)*vec(i) );
		}
		double f=0.0;
		for( unsigned int i = 0; i < exp.size(); ++i ) {
			f += 1.0 - exp(i);
		}
		ublas::vector<double> grad(exp);
		for( unsigned int i = 0; i < exp.size(); ++i ) {
			grad(i) = 2 * vec(i) * exp(i);
		}
		ublas::symmetric_matrix<double> hesse(vec.size(),vec.size());
		for( unsigned int i = 0; i < hesse.size1(); ++i ) {
			for( unsigned int j = 0; j < hesse.size2(); ++j ) {
				if( i != j ) {
					hesse(i,j) = 0.0;
				} else {
					hesse(i,j) = 2*exp(i)-4*vec(i)*vec(i)*exp(i);
				}
			}
		}
		
		return boost::make_tuple(f,grad,hesse);
	}
private:
	ublas::vector<double> center_;
};

struct gauss_least_squares {
	gauss_least_squares( ublas::vector<double>& center ): center_(center) {
	}

	boost::tuple< ublas::vector<double>, ublas::matrix<double>, ublas::symmetric_matrix<double> > operator() (const ublas::vector<double>& x) const {
		ublas::vector<double> vec = x - center_;
		ublas::vector<double> exp(vec);
		ublas::matrix<double> jacobian(exp.size(),exp.size());
		ublas::symmetric_matrix<double> hesse = ublas::zero_matrix<double>(exp.size(), exp.size());
		for( unsigned int i = 0; i < exp.size(); ++i ) {
			exp(i) = 1.0 - std::exp( -vec(i) );
		}
		for( unsigned int i = 0; i < jacobian.size1(); ++i ) {
			for( unsigned int j = 0; j < jacobian.size2(); ++j ) {
				if( i != j ) {
					jacobian(i,j) = 0.0;
				} else {
					jacobian(i,j) = std::exp( -vec(i) );
				}
			}
		}
		for( unsigned int k = 0; k < exp.size(); ++k ) {
		for( unsigned int i = 0; i < hesse.size1(); ++i ) {
			for( unsigned int j = 0; j < hesse.size2(); ++j ) {
				if( i != j ) {
					hesse(i,j) += 0.0;
				} else {
					hesse(i,j) += -exp(k)*std::exp( -vec(i) );
				}
			}
		}
		}
		return boost::make_tuple(exp,jacobian,hesse);
	}
private:
	ublas::vector<double> center_;
};

struct gauss_x_squared {
	gauss_x_squared( const ublas::vector<double>& center ): center_(center) {
	}
	
	boost::tuple< double, ublas::vector<double>, ublas::symmetric_matrix<double> > operator() (const ublas::vector<double>& x) const {
		ublas::vector<double> vec = x - center_;
		ublas::vector<double> exp(vec);
		for( unsigned int i = 0; i < exp.size(); ++i ) {
			exp(i) = std::exp( -vec(i)*vec(i) );
		}
		double fp=0.0;
		double fe=1.0;
		for( unsigned int i = 0; i < exp.size(); ++i ) {
			fp += vec(i)*vec(i);
			fe *= exp(i);
		}
		ublas::vector<double> grad(exp);
		for( unsigned int i = 0; i < exp.size(); ++i ) {
			grad(i) = 2*vec(i)*fe - 2*vec(i)*fp*fe;
		}
		ublas::symmetric_matrix<double> hesse(vec.size(),vec.size());
		for( unsigned int i = 0; i < hesse.size1(); ++i ) {
			for( unsigned int j = 0; j < hesse.size2(); ++j ) {
				hesse(i,j) = -4*vec(i)*vec(j)*(1.0-fp);
				hesse(i,j) += 2*(i!=j ? -2*vec(i)*vec(j) : 1.0 - fp -2*vec(i)*vec(j) );
				hesse(i,j) *= fe;
			}
		}
		
		return boost::make_tuple(fp*fe,grad,hesse);
	}
private:
	ublas::vector<double> center_;
};

#endif // _TEST_GAUSS_H
