#define BOOST_TEST_MODULE steihaug_toint
#include <boost/test/included/unit_test.hpp>

#include "gauss.h"
#include "rosenbrock.h"
#include "quadratic.h"

#include <steihaug_toint.h>

namespace detail {

	template<class T,class F> 
	boost::tuple< trm::stop::basic<T>, typename trm::type_traits<T>::function_type, typename trm::type_traits<T>::vector_type >
	steihaug_toint_regularization( const ublas::vector<T>& x0, F f, const trm::stop::basic<T>& stop_strategy, T regularization_alpha = 0.01 ) {
		trm::basic_driver< T, trm::delta::initial::nearest_cauchy,
			trm::core::steihaug_toint,
			trm::delta::update::simple_update,
			trm::stop::basic<T> > dr(x0,trm::functor::regularization<T>(f,trm::functor::vector_norm<T>(regularization_alpha, x0.size())),stop_strategy);
		return dr.minimize();
	}
	
}

template<class T,class F> void test_steihaug_toint( const ublas::vector<T>& initial_x, F function, T result, T abstol = 1e-6 ) {
	T function_min = 0.;
	boost::tie( boost::tuples::ignore, function_min, boost::tuples::ignore ) = trm::steihaug_toint(initial_x,function,trm::stop::basic<T>());

	BOOST_TEST_MESSAGE( "Function minimum: " << function_min );
	if( result != 0.0 ) {
		BOOST_CHECK_CLOSE( result, function_min, abstol );
	} else {
		BOOST_CHECK_SMALL( function_min, abstol );
	}
}

template<class T,class F> void test_steihaug_toint_nonnegative( const ublas::vector<T>& initial_x, F function, T result, T abstol = 1e-6 ) {
	T function_min = 0.;
	boost::tie( boost::tuples::ignore, function_min, boost::tuples::ignore ) = trm::steihaug_toint_nonnegative(initial_x,function,trm::stop::basic<T>());

	BOOST_TEST_MESSAGE( "Function minimum: " << function_min );
	if( result != 0.0 ) {
		BOOST_CHECK_CLOSE( result, function_min, abstol );
	} else {
		BOOST_CHECK_SMALL( function_min, abstol );
	}
}

template<class T,class F> void test_steihaug_toint_least_squares( const ublas::vector<T>& initial_x, F function, T result, T abstol = 1e-6 ) {
	T function_min = 0.;
	boost::tie( boost::tuples::ignore, function_min, boost::tuples::ignore ) = trm::steihaug_toint_least_squares(initial_x,function,trm::stop::basic<T>());

	BOOST_TEST_MESSAGE( "Function minimum: " << function_min );
	if( result != 0.0 ) {
		BOOST_CHECK_CLOSE( result, function_min, abstol );
	} else {
		BOOST_CHECK_SMALL( function_min, abstol );
	}
}

template<class T,class F> void test_steihaug_toint_nonnegative_least_squares( const ublas::vector<T>& initial_x, F function, T result, T abstol = 1e-6 ) {
	T function_min = 0.;
	boost::tie( boost::tuples::ignore, function_min, boost::tuples::ignore ) = trm::steihaug_toint_nonnegative_least_squares(initial_x,function,trm::stop::basic<T>());

	BOOST_TEST_MESSAGE( "Function minimum: " << function_min );
	if( result != 0.0 ) {
		BOOST_CHECK_CLOSE( result, function_min, abstol );
	} else {
		BOOST_CHECK_SMALL( function_min, abstol );
	}
}

template<class T,class F> void test_steihaug_toint_regularization( const ublas::vector<T>& initial_x, F function, T result, T norm, T abstol = 1e-6 ) {
	T function_min = 0.;
	ublas::vector<double> x;
	
	boost::tie( boost::tuples::ignore, function_min, x ) = detail::steihaug_toint_regularization(initial_x,function,trm::stop::basic<T>(),0.001);

	BOOST_TEST_MESSAGE( "Function minimum: " << function_min );
	BOOST_TEST_MESSAGE( "Result norm: " << ublas::norm_2(x) );
	if( result != 0.0 ) {
		BOOST_CHECK_CLOSE( result, function_min, abstol );
	} else {
		BOOST_CHECK_SMALL( function_min, abstol );
	}
	BOOST_CHECK_MESSAGE( ublas::norm_2(x) < norm, "result norm[" << ublas::norm_2(x) << "] not less than " << norm );
}

BOOST_AUTO_TEST_CASE( gauss_test ) {
	ublas::vector<double> x0(5);
	for( int i =0;i<x0.size();i++){
		x0(i)=1;
	}
	ublas::vector<double> center = 2 * x0;
	
	gauss g(center);
        test_steihaug_toint( x0,g , 0.0 );
}

BOOST_AUTO_TEST_CASE( rosenbrock_test ) {
	ublas::vector<double> x0(2);
	for( int i =0;i<x0.size();i++){
		x0(i)=1+i*2;
	}
	rosenbrock g;
        test_steihaug_toint( x0,g , 0.0 );
}

BOOST_AUTO_TEST_CASE( quadratic_test ) {
	ublas::vector<double> x0(2);
	x0(0)=1;
	x0(1)=1;
	ublas::vector<double> center;
	
	center = 2*x0;
	quadratic q1(center);
        test_steihaug_toint( x0,q1, 0.0 );
	
	center = -center;
	quadratic q2(center);
        test_steihaug_toint( x0,q2, 0.0 );

	center(0)=-center(0);
	quadratic q3(center);
        test_steihaug_toint( x0,q3, 0.0 );
	
	center = -center;
	quadratic q4(center);
        test_steihaug_toint( x0,q4, 0.0 );
}

BOOST_AUTO_TEST_CASE( gauss_test_nonnegative ) {
	ublas::vector<double> x0(5);
	for( int i =0;i<x0.size();i++){
		x0(i)=1;
	}
	ublas::vector<double> center = -x0;
	
	gauss g(center);
        test_steihaug_toint_nonnegative( x0,g , 3.1606027941427883 );
}

BOOST_AUTO_TEST_CASE( rosenbrock_test_nonnegative ) {
	ublas::vector<double> x0(2);
	for( int i =0;i<x0.size();i++){
		x0(i)=1+i*2;
	}
	rosenbrock g;
        test_steihaug_toint_nonnegative( x0,g , 0.0 );
}

BOOST_AUTO_TEST_CASE( quadratic_test_nonnegative ) {
	ublas::vector<double> x0(2);
	x0(0)=1;
	x0(1)=1;
	ublas::vector<double> center;
	
	center = 2*x0;
	quadratic q1(center);
        test_steihaug_toint_nonnegative( x0,q1, 0.0 );
	
	center = -center;
	quadratic q2(center);
        test_steihaug_toint_nonnegative( x0,q2, 4*2.0 );

	center(0)=-center(0);
	quadratic q3(center);
        test_steihaug_toint_nonnegative( x0,q3, 4*1.0 );
	
	center = -center;
	quadratic q4(center);
        test_steihaug_toint_nonnegative( x0,q4, 4*1.0 );
}

BOOST_AUTO_TEST_CASE( gauss_test_least_squares ) {
	ublas::vector<double> x0(5);
	for( int i =0;i<x0.size();i++){
		x0(i)=1;
	}
	ublas::vector<double> center = 2 * x0;
	
	gauss_least_squares g(center);
        test_steihaug_toint_least_squares( x0,g , 0.0 );
}

BOOST_AUTO_TEST_CASE( gauss_test_nonnegative_least_squares ) {
	ublas::vector<double> x0(5);
	for( int i =0;i<x0.size();i++){
		x0(i)=1;
	}
	ublas::vector<double> center = 2 * x0;
	
	gauss_least_squares g(center);
        test_steihaug_toint_nonnegative_least_squares( x0,g , 0.0 );
}

BOOST_AUTO_TEST_CASE( gauss_x_squared_test ) {
	ublas::vector<double> x0(5);
	ublas::vector<double> center(x0.size());
	for( int i=0;i<x0.size();i++ ){
		x0(i)=2.0;
		center(i)=0;
	}
	
	gauss_x_squared g(center);
        test_steihaug_toint_regularization( x0,g , 0.0, 1.0 );
}